import os
import re
from datetime import datetime
from itertools import takewhile

import numpy as np
import pint_xarray

from peaks.core.fileIO.base_arpes_data_classes.base_arpes_data_class import (
    BaseARPESDataLoader,
)
from peaks.core.fileIO.loc_registry import register_loader
from peaks.core.utils.misc import analysis_warning

ureg = pint_xarray.unit_registry


@register_loader
class SpecsDataLoader(BaseARPESDataLoader):
    """Data loader for Specs ARPES data measured using Specs Prodigy.

    Define _scan_axis_resolution_order to define preferences for the primary dimension of
    a 3D scan where more than one user axis varies."""

    _loc_name = "Specs"
    _loc_description = "SPECS Phoibos Analysers with Prodigy control"
    _loc_url = "https://www.specs-group.com/"
    _analyser_slit_angle = None
    _scan_axis_resolution_order = []
    _SPECS_metadata_key_mappings = {}
    _SPECS_metadata_units = {}
    _analyser_include_in_metadata_warn = [
        "eV",
        "PE",
        "sweeps",
        "dwell",
        "azi",
    ]

    @classmethod
    def _load_data(cls, fpath, lazy):
        data_parsers = {
            ".xy": cls._parse_data_from_xy_file,
            ".sp2": cls._parse_data_from_sp2_file,
        }
        file_extension = os.path.splitext(fpath)[-1]
        if file_extension in data_parsers:
            return data_parsers[file_extension](fpath)
        else:
            raise ValueError(
                f"File extension {file_extension} not supported for data parsing."
            )

    @classmethod
    def _parse_data_from_xy_file(cls, fpath):
        # Open the file and load lines
        with open(fpath) as f:
            lines = f.readlines()
        # Split into metadata-type and data
        meta_lines = [line for line in lines if line.startswith("#")]
        data = np.loadtxt(lines)

        # Parse the metadata, including extracting any user-dependent axis variations
        metadata_dict = cls._parse_metalines(meta_lines)

        # Parse the energies and counts
        counts = data[:, 1]
        energies = data[:, 0]
        # Extract the non-repeating KE values
        start_KEs = np.where(energies == energies[0])
        number_of_KEs = start_KEs[0][1]
        KE_values = energies[0:number_of_KEs]
        metadata_dict["eV scale"] = (
            KE_values  # Add to metadata dict for use when caching
        )

        # Cache the metadata to use later if required
        cls._metadata_cache[fpath] = metadata_dict
        if "scan_parameters" in metadata_dict:
            parameters_dict = metadata_dict.pop("scan_parameters")
        if metadata_dict.get("Count Rate") == "Counts per Second":
            spectrum_units = "counts/s"
        else:
            spectrum_units = "counts"

        # Extract the theta_par values
        # (only need to take from the first cycle as they repeat after)
        cycle0_lines = list(
            takewhile(lambda line: not line.startswith("# Cycle: 1"), meta_lines)
        )
        theta_par_values = [
            float(line.split(": ")[1])
            for line in cycle0_lines
            if line.startswith("# NonEnergyOrdinate:")
        ]

        if parameters_dict:
            # Try and parse the names from local names to peaks names
            inverse_manipulator_name_mappings = {
                v: k for k, v in cls._manipulator_name_conventions.items()
            }
            for key in list(parameters_dict.copy().keys()):
                peaks_name = inverse_manipulator_name_mappings.get(key[0], key[0])
                parameters_dict[(peaks_name, key[1])] = parameters_dict.pop(key)

        # Reshape the data
        if len(parameters_dict) == 1:
            user_axis, values = next(iter(parameters_dict.items()))
            spectrum = counts.reshape(
                (len(values), len(theta_par_values), number_of_KEs)
            )
            return {
                "spectrum": spectrum,
                "dims": [user_axis[0], "theta_par", "eV"],
                "coords": {
                    user_axis[0]: values,
                    "theta_par": theta_par_values,
                    "eV": KE_values,
                },
                "units": {
                    "eV": "eV",
                    "theta_par": "deg",
                    user_axis[0]: user_axis[1],
                    "spectrum": spectrum_units,
                },
            }
        elif len(parameters_dict) == 0:
            spectrum = counts.reshape((len(theta_par_values), number_of_KEs))
            return {
                "spectrum": spectrum,
                "dims": ["theta_par", "eV"],
                "coords": {"theta_par": theta_par_values, "eV": KE_values},
                "units": {"eV": "eV", "theta_par": "deg", "spectrum": spectrum_units},
            }
        else:
            # If all the user axes have the same range, then we can parse this still
            # with secondary coords
            if len(set(len(v) for v in parameters_dict.values())) == 1:
                user_axes = list(parameters_dict.keys())
                user_axis_names = [axis[0] for axis in user_axes]
                user_axis_dimensions = [axis[1] for axis in user_axes]
                user_axis_len = len(parameters_dict[user_axes[0]])
                spectrum = counts.reshape(
                    (
                        user_axis_len,
                        len(theta_par_values),
                        number_of_KEs,
                    )
                )
                # Check if a resolution order for these axes has been defined to
                # determine the primary dim
                primary_dim = None
                if cls._scan_axis_resolution_order:
                    for dim in cls._scan_axis_resolution_order:
                        if dim.lower() in [name.lower() for name in user_axis_names]:
                            primary_dim = dim
                            break
                if not primary_dim:
                    primary_dim = user_axis_names[0]
                primary_dim_label = user_axes[user_axis_names.index(primary_dim)]

                # Primary coords
                coords = {
                    "theta_par": theta_par_values,
                    "eV": KE_values,
                    primary_dim: parameters_dict[primary_dim_label],
                }
                # Secondary coords
                for axis in user_axes:
                    if axis != primary_dim_label:
                        coords[axis[0]] = (primary_dim, parameters_dict[axis])
                return {
                    "spectrum": spectrum,
                    "dims": [primary_dim, "theta_par", "eV"],
                    "coords": coords,
                    "units": {
                        "eV": "eV",
                        "theta_par": "deg",
                        **{
                            user_axis_names[i]: user_axis_dimensions[i]
                            for i in range(len(user_axis_names))
                        },
                        "spectrum": spectrum_units,
                    },
                }
            else:
                raise ValueError(
                    "Multiple user-varying parameters detected. Scan type not currently "
                    "supported by loader."
                )

    @classmethod
    def _parse_data_from_sp2_file(cls, fpath):
        metadata_dict_SPECS_keys = cls._load_metadata(fpath, return_in_SPECS_format=True)

        # Open the file and load lines
        rows = []
        with open(fpath, "rb") as f:
            for i, row in enumerate(f):  # noqa: B007
                if i >= metadata_dict_SPECS_keys["data_start_line"]:
                    rows.append(int(row.decode(errors="ignore").strip("\n")))

        data = np.asarray(rows).reshape(
            int(metadata_dict_SPECS_keys["SIZE_Y"].split("#")[0]),
            int(metadata_dict_SPECS_keys["SIZE_X"].split("#")[0]),
        )

        # Extract the scale values
        x0, x1, x_units = cls._extract_numbers_and_unit_from_range(
            metadata_dict_SPECS_keys["X Range"]
        )
        y0, y1, y_units = cls._extract_numbers_and_unit_from_range(
            metadata_dict_SPECS_keys["Y Range"]
        )
        return {
            "spectrum": data,
            "dims": [
                "y",
                "eV",
            ],
            "coords": {
                "eV": np.linspace(x0, x1, data.shape[1]),
                "y": np.linspace(y0, y1, data.shape[0]),
            },
            "units": {"eV": x_units, "y": y_units, "spectrum": "counts"},
        }

    @staticmethod
    def _extract_calib2D_info(metadata_dict_SPECS_keys):
        """Extract the calibration information from the metadata."""
        # ToDo: Complete this method and implement transformation logic
        # Extract the calib2d info from the metadata
        calib2d = {}
        calib2d["eRange"] = [
            float(metadata_dict_SPECS_keys["ERange"].split(" ")[0]),
            float(metadata_dict_SPECS_keys["ERange"].split(" ")[1]),
        ]
        calib2d["De1"] = float(metadata_dict_SPECS_keys["De1"].split(" ", 1)[0])
        calib2d["eShift"] = (
            metadata_dict_SPECS_keys.get("EShift").split("#")[0].strip().split(" ")
        )
        calib2d["eShift"] = [float(val) for val in calib2d["eShift"]]
        calib2d["aRange"] = [
            float(metadata_dict_SPECS_keys["ARange"].split(" ")[0]),
            float(metadata_dict_SPECS_keys["ARange"].split(" ")[1]),
        ]
        calib2d["aInner"] = float(metadata_dict_SPECS_keys["aInner"].split(" ")[0])

    @staticmethod
    def _extract_numbers_and_unit_from_range(s):
        pattern = r"\[([-+]?\d*\.\d+|\d+) \.\. ([-+]?\d*\.\d+|\d+)\] (\w+)"
        match = re.search(pattern, s)
        if match:
            num1 = float(match.group(1))
            num2 = float(match.group(2))
            unit = match.group(3)
            return num1, num2, unit
        else:
            raise ValueError("String format is incorrect")

    @classmethod
    def _load_metadata(cls, fpath, return_in_SPECS_format=False):
        """Load metadata from an Prodigy file."""
        # Check if there is a cached version (cache with keys in SPECS format plus a
        # "scan_parameters" key with custom axis metadata)
        metadata_dict_SPECS_keys = cls._metadata_cache.get(fpath)

        # If no metadata in the cache, load it
        if not metadata_dict_SPECS_keys:
            metadata_parsers = {
                ".xy": cls._parse_metadata_from_xy_file,
                ".sp2": cls._parse_metadata_from_sp2_file,
            }
            file_extension = os.path.splitext(fpath)[-1]
            if file_extension in metadata_parsers:
                metadata_dict_SPECS_keys = metadata_parsers[file_extension](fpath)
            else:
                raise ValueError(
                    f"File extension {file_extension} not supported for metadata parsing"
                )

        if return_in_SPECS_format:
            return metadata_dict_SPECS_keys

        # Convert the metadata to the peaks convention and return
        return cls._SPECS_metadata_dict_keys_to_peaks_keys(metadata_dict_SPECS_keys)

    @classmethod
    def _parse_metadata_from_xy_file(cls, fpath):
        with open(fpath) as f:
            lines = f.readlines()
        metadata_dict_SPECS_keys = cls._parse_metalines(
            [line for line in lines if line.startswith("#")]
        )
        # Get energy axis info to add to the file
        energy_lines = list(takewhile(lambda line: line.startswith("#"), lines))
        eV = [line.split(" ") for line in energy_lines if not line.startswith("#")]
        metadata_dict_SPECS_keys["eV scale"] = eV[0][0]

        return metadata_dict_SPECS_keys

    @classmethod
    def _parse_metadata_from_sp2_file(cls, fpath):
        meta_dict_SPECS_keys = {}
        stop_on_next_line = False  # noqa: F841
        with open(fpath, "rb") as f:
            for i, row in enumerate(f):  # noqa: B007
                row_ = row.decode(errors="ignore")
                if isinstance(row_, str) and "=" in row_:
                    meta = row_.split("=", 1)
                    meta_dict_SPECS_keys[meta[0].split("#", 1)[1].strip()] = meta[
                        1
                    ].strip()
                    if meta[0].strip() == "SIZE_X":
                        x_pixels = int(meta[1].strip())  # noqa: F841
                if (
                    meta_dict_SPECS_keys.get("SIZE_X") is not None
                    and meta_dict_SPECS_keys.get("SIZE_X").split("#")[0].strip()
                    == row_.split(" ")[0].strip()
                ):
                    break
            meta_dict_SPECS_keys["data_start_line"] = i + 1
        return meta_dict_SPECS_keys

    @classmethod
    def _parse_metalines(self, meta_lines):
        # Iterate over the first part of the metadata file up to where the data starts
        initial_metadata_lines = list(
            takewhile(
                lambda line: not line.startswith("# NonEnergyOrdinate"), meta_lines
            )
        )
        meta_dict = {
            line.split(":", 1)[0].split("#")[1].strip(): line.split(":", 1)[1].strip()
            for line in initial_metadata_lines
            if len(line.split(":")) > 1
        }

        # Extract any user-varying parameter (e.g. for a FS map)
        parameter_regex = re.compile(
            r'# Parameter: "([^"]+)\s*\[(\w+)\]" = (-?\d+\.?\d*)'
        )  # Compile regex pattern
        parameters_dict = {}
        for line in meta_lines:
            if match := parameter_regex.match(line):
                name, dim, value = (
                    match.group(1).strip(),
                    match.group(2).strip(),
                    float(match.group(3)),
                )
                key = (name, dim)
                if key not in parameters_dict:
                    parameters_dict[key] = []
                parameters_dict[key].append(value)

        meta_dict["scan_parameters"] = parameters_dict

        return meta_dict

    @classmethod
    def _SPECS_metadata_dict_keys_to_peaks_keys(cls, metadata_dict_SPECS_keys):
        """Extract metadata values in peaks conventions and assign units where needed.

        Parameters
        ------------
        metadata_dict_SPECS_keys : dict
            Dictionary of metadata key-value pairs with keys in SPECS format.

        Returns
        ------------
        metadata_dict : dict
            Dictionary of metadata key-value pairs with keys in peaks format.

        Notes
        ------------
        The extraction process is based on the mappings defined in the dictionaries
        `standard_keys` and `standard_units`. The entries of these dictionaries are
        updated from the class variables `_SPECS_metadata_key_mappings` and
        `_SPECS_metadata_units` respectively, and so subclasses can overwrite and extend
        these defaults by specifying the appropriate mappings in these class variables.

        The extraction process supports different types of keys:
            - If a single key is given, the value is extracted directly.
            - If a list of keys is given, the function tries to extract values for all
            keys and returns a list or array of all non-None values.
            - If a callable is given, the function calls it, passing it the metadata
            dictionary and returns the result.
            - A constant value can be given by defining a simple lambda function, for
            example `lambda x: -1`.

        """

        def _parse_timestamp(metadata_dict_SPECS_keys):
            """Parse the timestamp from the metadata."""
            # Extract the date and time from the metadata
            date_string = metadata_dict_SPECS_keys.get("Acquisition Date")
            if date_string is None:
                return None
            date_format = "%m/%d/%y %H:%M:%S %Z"
            parsed_date = datetime.strptime(date_string, date_format)
            formatted_date = parsed_date.strftime("%Y-%m-%d %H:%M:%S")
            return formatted_date

        def _parse_analyser_slit_width(metadata_dict_SPECS_keys):
            slit_info = metadata_dict_SPECS_keys.get("Analyzer Slit")
            if slit_info and isinstance(slit_info, str):
                try:
                    return try_parse_to_float(slit_info.split(":")[1].split("x")[0])
                except (IndexError, AttributeError):
                    return None

        def _parse_energy_scale(metadata_dict_SPECS_keys):
            """Parse the energy scale from the metadata."""
            # Extract the energy scale from the metadata
            energy_scale = metadata_dict_SPECS_keys.get("eV scale")
            try:
                return [min(energy_scale), max(energy_scale)]
            except (TypeError, ValueError):
                return None

        def _parse_energy_step(metadata_dict_SPECS_keys):
            """Parse the energy step from the metadata."""
            # Extract the energy step from the metadata
            energy_scale = metadata_dict_SPECS_keys.get("eV scale")
            try:
                return np.ptp(energy_scale) / (len(energy_scale) - 1)
            except (TypeError, ValueError):
                return None

        # Define the mapping assuming the standard keys for SES metadata
        standard_keys = {
            "analyser_model": "Analyzer",
            "analyser_slit_width": _parse_analyser_slit_width,
            "analyser_slit_width_identifier": "Analyzer Slit",
            "analyser_eV": lambda x: (
                try_parse_to_float(x.get("Kinetic Energy"))
                if x.get("Scan Mode") == "SnapshotFAT"
                else _parse_energy_scale(x)
            ),
            "analyser_eV_type": "Energy Axis",
            "analyser_step_size": lambda x: (
                _parse_energy_step(x) if x.get("Scan Mode") != "SnapshotFAT" else None
            ),
            "analyser_PE": "Pass Energy",
            "analyser_sweeps": "Number of Scans",
            "analyser_dwell": "Dwell Time",
            "analyser_lens_mode": "Analyzer Lens",
            "analyser_acquisition_mode": "Scan Mode",
            "analyser_polar": None,
            "analyser_tilt": None,
            "analyser_azi": None,
            "analyser_deflector_parallel": None,
            "analyser_deflector_perp": None,
            "timestamp": _parse_timestamp,
            "manipulator_polar": cls._manipulator_name_conventions.get("polar", None),
            "manipulator_tilt": cls._manipulator_name_conventions.get("tilt", None),
            "manipulator_azi": cls._manipulator_name_conventions.get("azi", None),
            "manipulator_x1": cls._manipulator_name_conventions.get("x1", None),
            "manipulator_x2": cls._manipulator_name_conventions.get("x2", None),
            "manipulator_x3": cls._manipulator_name_conventions.get("x3", None),
            "photon_hv": "Excitation Energy",
        }
        # Define standard units
        standard_units = {
            "analyser_model": None,
            "analyser_slit_width": "mm",
            "analyser_slit_width_identifier": None,
            "analyser_eV": "eV",
            "analyser_step_size": "eV",
            "analyser_PE": "eV",
            "analyser_sweeps": None,
            "analyser_dwell": "s",
            "analyser_lens_mode": None,
            "analyser_acquisition_mode": None,
            "analyser_polar": "deg",
            "analyser_tilt": "deg",
            "analyser_azi": "deg",
            "photon_hv": "eV",
        }

        # If custom key mappings or units have been supplied, update the standard keys
        standard_keys.update(cls._SPECS_metadata_key_mappings)
        standard_units.update(cls._SPECS_metadata_units)

        # Extract metadata values and give them units where appropriate
        # - if a single key is given, extract the value directly
        # - if a list of keys is given, try and extract all and return non-None values
        # - if a callable is given, call the function with the metadata dictionary and
        # return the result
        # - add units if given and extracted value is not None
        metadata_dict = {}
        units_failure = []

        def try_parse_to_float(value):
            try:
                return float(value)
            except (ValueError, TypeError):
                return value

        # Handle extracting metadata based on arguments in the standard_keys dictionary
        for peaks_key, local_key in standard_keys.items():
            value = None
            if isinstance(local_key, str):
                value = try_parse_to_float(metadata_dict_SPECS_keys.get(local_key))
            elif callable(local_key):
                value = local_key(try_parse_to_float(metadata_dict_SPECS_keys))
            elif isinstance(local_key, list):
                value = [
                    try_parse_to_float(metadata_dict_SPECS_keys.get(key))
                    for key in local_key
                    if metadata_dict_SPECS_keys.get(key) is not None
                ]

                if len(value) == 1:
                    value = value[0]
                elif len(value) == 0:
                    value = None
                else:
                    try:
                        value = np.asarray(value)
                    except (ValueError, TypeError):
                        value = str(value)

            # Try to add units if they are given and the value is not None
            if value is not None and standard_units.get(peaks_key):
                try:
                    value = value * ureg(standard_units[peaks_key])
                except TypeError:
                    units_failure.append(peaks_key)  # Keep track of errors
            metadata_dict[peaks_key] = value

        # Warn if any units failed to be added
        if units_failure:
            analysis_warning(
                f"Failed to add units to the following metadata items: {units_failure}",
                "warning",
                "Failed to add units to metadata",
            )

        return metadata_dict



# ruff: noqa

"""peaks: (P)ython (E)lectron Spectroscopy and Diffraction (A)nalysis by (K)ing Group (S)t Andrews.

peaks is a collection of analysis tools for the loading, processing and display of spectroscopic and diffraction data,
with a core focus on tools for angle-resolved photoemission and electron diffraction techniques.

Usage
----------
It is recommended to import peaks as follows::

    import peaks as pks

To load data, use the load function::

    data = pks.load('data.ibw')

"""

__version__ = "0.4.9"

# Set some default xarray options
import xarray as xr

# Set default xarray options
xr.set_options(
    cmap_sequential="binary",
    use_numbagg=True,
    display_expand_attrs=False,
    display_expand_data=False,
)

# # Register a progressbar for use during dask compute calls
from peaks.core.utils.misc import DaskTQDMProgressBar

DaskTQDMProgressBar(desc="Evaluating deferred computation", minimum=1.0).register()

# Enable pint accessor and set default options
import pint_xarray

ureg = pint_xarray.unit_registry
ureg.formatter.default_format = (
    "~P"  # Set formatting option to short form (with symbols)
)

# Register the relevant data loaders and load method (do this first to get the loc_registry populated)
from peaks.core.fileIO.loc_registry import LOC_REGISTRY
from peaks.core.fileIO.loc_registry import locs
from peaks.core.fileIO.base_data_classes.base_ibw_class import BaseIBWDataLoader
from peaks.core.fileIO.base_arpes_data_classes.base_ses_class import SESDataLoader
from peaks.core.fileIO.base_arpes_data_classes.base_fesuma_class import (
    BaseFeSuMaDataLoader,
)
from peaks.core.fileIO.base_arpes_data_classes.base_mbs_class import MBSDataLoader
from peaks.core.fileIO.loaders import *

# Import the core functions that should be accessible from the main peaks namespace
from peaks.core.fileIO.data_loading import load
from peaks.core.fitting.fit import load_fit
from peaks.core.options import opts
from peaks.core.display.plotting import (
    plot_grid,
    plot_DCs,
    plot_ROI,
    plot_kpar_cut,
    plot_kz_cut,
    plot_nanofocus,
)
from peaks.core.process.tools import sum_data, subtract_data, merge_data

# Register the relevant accessor functions
from peaks.core.accessors import *



import h5py
import numpy as np
import pint
from termcolor import colored

from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader, ureg


class BaseHDF5DataLoader:
    """Helper class for automating parsing data and metadata from hdf5 files.

    Notes
    -----
    Subclasses should define the _hdf5_metadata_key_mappings and
    _hdf5_metadata_fixed_units class attributes.
    The extraction process supports different types of keys:
        - If a single key is given, the value is attempted to be extracted directly
        using that key.
        - If a `list` or `tuple` of keys is given, the function tries to extract values
        for each key in turn, returning
        the first value which returns a result.
        - If a callable is given, the function is passed the h5py file object. It should
        return either a key (string)
        or directly return a value (float, int etc.).
        - A constant value (:class:`pint.Quantity`, `float`, `int`) can be provided by
        giving that value directly.
        - A constant string can be given by passing a string that starts with
        "FIXED_VALUE:", e.g. "FIXED_VALUE:LH".

    Note, if a key is passed as a string, units will attempt to be parsed from the
    relevant hdf5 dataset attributes. If units are present in the file, this takes
    precedence over any fixed units defined in the `_hdf5_metadata_fixed_units` class
    attribute. If units are not present in the file, the fixed units will be used if
    supplied. If passing a callable in the metadata mapping, the recommended approach is
    to use the function to parse the relevant key to call, and then finally pass the key
    to the `_extract_hdf5_value` method to extract the value from the hdf5 file.
    In this way, consistent handling of units is ensured.
    """

    # mappings from desired metadata keys to hdf5 field addresses
    _hdf5_metadata_key_mappings = {}
    # mappings from metadata keys to fixed units,
    # otherwise these will be attempted to be determined from the hdf5 field attributes
    _hdf5_metadata_fixed_units = {}

    @classmethod
    def _make_dataarray(cls, data):
        """Ensure data is in a :class:`xarray.DataArray` format for HDF5 loaders, which
        typically already convert to :class:`xarray.DataArray` in `_load_data` method.

        Parameters
        ----------
        data : dict or xarray.DataArray
            If a dictionary, follow convention in :meth:`BaseDataLoader._load_data`.
        """
        if isinstance(data, dict):
            return BaseDataLoader._make_dataarray(data)
        return data

    @classmethod
    def _load_metadata(cls, fpath):
        """Load metadata from a Diamond hdf5 file.

        Parameters
        ----------
        fpath : str
            Path to the file to be loaded.

        Returns
        -------
        metadata : dict
            Dictionary containing the extracted metadata.

        """
        # Open the file (read only)
        with h5py.File(fpath, "r") as f:
            metadata = {}
            for (
                peaks_key,
                hdf5_key,
            ) in cls._hdf5_metadata_key_mappings.items():
                if isinstance(hdf5_key, (list, tuple)):
                    # Iterate through all keys until a value is returned
                    metadata_entry = None
                    for key in hdf5_key:
                        metadata_entry = cls._extract_hdf5_value(f, key)
                        if metadata_entry is not None:
                            break
                    metadata[peaks_key] = metadata_entry
                elif hdf5_key is not None:
                    metadata[peaks_key] = cls._extract_hdf5_value(f, hdf5_key)
                else:
                    metadata[peaks_key] = None
            return metadata

    @classmethod
    def _extract_hdf5_value(cls, f, key, return_extreme_values=True):
        """Extract a value from a hdf5 file, adding units if possible.

        Parameters
        ----------
        f : h5py.File
            The h5py file object (should be open).
        key : pint.Quantity, str, float, int, list, np.ndarray, callable
            The key to extract from the file.
        return_extreme_values : bool, optional
            Whether to return the extreme values of an array if it is an array or list.
            Defaults to True. If False, returns entire array.

        Returns
        -------
        extracted_value : Union[Quantity, str, float, int, list, np.ndarray, None]
            The extracted value. Preference given to pint.Quantity's where possible.

        """
        if isinstance(key, (float, int, list, np.ndarray, pint.Quantity)):
            return key
        elif callable(key):
            # Call the function and pass it back here.
            # This works if a key returned as a string from the function, otherwise also
            # to check for units
            return cls._extract_hdf5_value(f, key(f))
        elif isinstance(key, str):
            # Check for a direct string flag
            if key.startswith("FIXED_VALUE:"):
                return key.split("FIXED_VALUE:")[1]
            # Otherwise this should be a key. Try and extract value from the hdf5 file
            try:
                value = f[key][()]
                value = value.decode() if isinstance(value, bytes) else value
                if isinstance(value, (np.ndarray, list)):
                    if len(value) == 1:
                        while isinstance(value, (np.ndarray, list)):
                            value = value[0]
                    elif return_extreme_values:
                        value = np.array([np.min(value), np.max(value)])
                        if value[0] == value[-1]:
                            value = value[0]

                value = value.decode() if isinstance(value, bytes) else value
                units = f[key].attrs.get("units")
                units = (
                    units[0]
                    if isinstance(units, (np.ndarray, list)) and len(units) == 1
                    else units
                )
                units = (
                    units.decode() if isinstance(units, (bytes, np.bytes_)) else units
                )
                if not units:
                    # If can't parse from the file, check for a fixed units string
                    units = cls._hdf5_metadata_fixed_units.get(key)

                return value * ureg(units) if units else value
            except KeyError:
                return None
        if key is None:
            return None
        else:
            raise ValueError(
                f"Invalid key in metadata mapping key: {type(key)}. "
                f"Expected pint.Quantity, str, float, int, or callable."
            )

    @classmethod
    def _print_hdf5_structure(
        cls, name, obj, parent_group, indent_level=0, is_last=False, branch=""
    ):
        """Recursive function to print the structure of the HDF5 file with colored keys
        and default (black) data, and indentation lines.

        Parameters
        ----------
        name : str
            The name of the current object.

        obj : h5py.Group, h5py.Dataset, h5py.SoftLink, h5py.ExternalLink
            The object to explore.

        parent_group : h5py.Group
            The parent group of the current object.

        indent_level : int, optional
            The current indentation level of the object.

        is_last : bool, optional
            Whether the current object is the last child of the parent group.

        branch : str, optional
            The current branch of the object.

        """
        # Color definitions
        group_color = "cyan"
        dataset_color = "yellow"
        attribute_color = "green"
        link_color = "magenta"

        # Set the line connector based on whether the node is the last child
        connector = "└── " if is_last else "├── "

        # Build the branch line for current level
        new_branch = branch + ("    " if is_last else "│   ")

        if isinstance(obj, h5py.Group):
            # Print groups in cyan
            group_name = colored(
                f"{name.split('/')[-1]}:NX{obj.attrs.get('NX_class', 'unknown')}",
                group_color,
            )
            print(f"{branch}{connector}{group_name}")

            # Recursively explore the contents of the group
            keys = list(obj.keys())
            for i, key in enumerate(keys):
                sub_obj = obj[key]
                cls._print_hdf5_structure(
                    key,
                    sub_obj,
                    obj,
                    indent_level + 1,
                    is_last=(i == len(keys) - 1),
                    branch=new_branch,
                )

        elif isinstance(obj, h5py.Dataset):
            # Print datasets in yellow, but display data in default black if large data
            # is compacted
            data_info = (
                f"shape={obj.shape}, dtype={obj.dtype}"
                if obj.size > 10
                else f"{obj[()]}"
            )
            dataset_name = colored(f"{name.split('/')[-1]} =", dataset_color)
            print(f"{branch}{connector}{dataset_name} {data_info}")

            # Print dataset attributes in green, display attribute values in black
            for attr_name, attr_value in obj.attrs.items():
                attr_display = colored(f"@{attr_name} =", attribute_color)
                print(f"{new_branch}  {attr_display} {attr_value}")

        elif isinstance(obj, h5py.SoftLink) or isinstance(obj, h5py.ExternalLink):
            # Handle links
            link_name = colored(
                f"{name.split('/')[-1]} -> {obj.path} (Link)", link_color
            )
            print(f"{branch}{connector}{link_name}")

    @classmethod
    def hdf5_explorer(cls, file_path):
        """Function to explore the structure of an HDF5 file, printing the keys, groups,
        datasets and attributes.

        Parameters
        ----------
        file_path : str
            The path to the HDF5 file to explore.

        Examples
        --------
        Example usage is as follows::

            import peaks as pks

            # Explore the structure of an HDF5 file
            pks.hdf5_explorer('data.h5')

        Notes
        -----
        Colored output to distinguish between groups, datasets, attributes and links:
            - Cyan: For groups (NX classes like NXentry, NXdata, etc.)
            - Yellow: For datasets (the actual data arrays)
            - Green: For attributes (metadata attached to datasets or groups, typically
            prefixed with @)
            - Magenta: For soft links (references to other datasets/groups within file)
            - Black: For data values (the actual contents of datasets and attributes)

        Note, soft links may resolve automatically and so not show as links in the output
        """
        with h5py.File(file_path, "r") as f:
            # Manually explore the root level items
            keys = list(f.keys())
            for i, key in enumerate(keys):
                obj = f[key]
                cls._print_hdf5_structure(key, obj, f, is_last=(i == len(keys) - 1))



import pint_xarray

ureg = pint_xarray.unit_registry


def _set_t0(da, t0, delay_line_no_trips=2, assign=True):
    """Set a new t0 in a TR-ARPES data set

    Parameters
    -----------
    da : xarray.DataArray
        da : xarray.DataArray
        time-resolved data, with a time axis with dimension "t", and a metadata attribute pump.t0_position defining
        the delay line position corresponding to t0.

    t0 : pint.Quantity or float
        New time zero, expressed as a time based on the current time axis scaling.

    delay_line_no_trips : int, optional
        Number of times delay line is traversed. Default is 2 (there and back).

    assign : bool, optional
        If True, the new t0 is assigned to the :class:xarray.DataArray.
        If False, the :class:xarray.DataArray is updated in place

    Retruns
    -------
    xarray.DataArray : optional
        If assign is False, the updated xarray.DataArray is returned.

    See Also
    --------
    set_t0
    assign_t0
    """

    if not t0:  # No correction to do if t0 = 0
        return da

    original_t_units = da.t.units
    original_stage_units = da.metadata.pump.t0_position.units
    delay_position = da.delay_pos.copy(deep=True)

    if isinstance(t0, (float, int)):
        t0 = t0 * original_t_units

    # Get current t0 in stage position
    t0_old_position = da.metadata.pump.t0_position
    if t0.check("[length]"):  # If provided as a length, set that directly
        t0_new_position = t0
    else:  # Otherwise convert the given t0 time into a stage position
        # Work out different in stage position for new t0
        t0_delta = (t0 / (delay_line_no_trips / ureg.c)).to(original_stage_units)
        t0_new_position = t0_old_position + t0_delta
    # Convert to time in original units
    delay_time = (
        (delay_position - t0_new_position) * delay_line_no_trips / ureg.c
    ).pint.to(original_t_units)
    hist_str = (
        f"New t0 defined as {t0}. Stage position corresponding to t0 updated - "
        f"old: {t0_old_position}, new: {t0_new_position}"
    )

    if assign:  # Assign rather than update in place
        # Make new da with updated delay dim
        new_da = da.copy(deep=True).assign_coords(
            {"t": delay_time.pint.dequantify(), "delay_pos": ("t", delay_position.data)}
        )
        new_da.t.attrs["units"] = delay_time.data.units
        # Update the metadata
        new_da.metadata.pump.set("t0_position", t0_new_position, add_history=False)
        new_da.history.add(hist_str)
        return new_da

    # Update the metadata in place
    da.metadata.pump.set("t0_position", t0_new_position, add_history=False)
    da.history.add(hist_str)
    # Update the time axis
    da["t"] = delay_time.data.magnitude
    da.t.attrs["units"] = delay_time.data.units
    da["delay_pos"] = ("t", delay_position.data)


def set_t0(da, t0, delay_line_roundtrips=2):
    """Set a new t0 for a time-resolved data set

    Parameters
    -----------
    da : xarray.DataArray
        da : xarray.DataArray
        time-resolved data, with a time axis with dimension "t", and a metadata attribute pump.t0_position defining
        the delay line position corresponding to t0.

    t0 : pint.Quantity or float
        New time zero, expressed as a time based on the current time axis scaling.

    delay_line_roundtrips : int, optional
        Number of round trips in the delay line. Default is 2.

    Retruns
    -------
    None
        The :class:xarray.DataArray is updated in place
    """

    _set_t0(da, t0, delay_line_roundtrips, assign=False)


def set_t0_like(da, da_ref):
    """Set t0 of a time-resolved data set to match another data set. Assumes the default delay line round trips of 2.

    Parameters
    -----------
    da : xarray.DataArray
        time-resolved data, with a time axis with dimension "t", and a metadata attribute pump.t0_position defining
        the delay line position corresponding to t0.

    da_ref : xarray.DataArray
        Reference time-resolved data set, with the t0 already calibtrated
    """

    t0 = da_ref.metadata.pump.t0_position
    set_t0(da, t0)


def assign_t0(da, t0, delay_line_roundtrips=2):
    """Set a new t0 in a TR-ARPES data set

    Parameters
    -----------
    da : xarray.DataArray
        da : xarray.DataArray
        time-resolved data, with a time axis with dimension "t", and a metadata attribute pump.t0_position defining
        the delay line position corresponding to t0.

    t0 : pint.Quantity or float
        New time zero, expressed as a time based on the current time axis scaling.

    delay_line_roundtrips : int, optional
        Number of round trips in the delay line. Default is 2.

    Retruns
    -------
    xarray.DataArray
        The updated xarray.DataArray
    """

    return _set_t0(da, t0, delay_line_roundtrips, assign=True)


def assign_t0_like(da, da_ref):
    """Set t0 of a time-resolved data set to match another data set. Assumes the default delay line round trips of 2.

    Parameters
    -----------
    da : xarray.DataArray
        time-resolved data, with a time axis with dimension "t", and a metadata attribute pump.t0_position defining
        the delay line position corresponding to t0.

    da_ref : xarray.DataArray
        Reference time-resolved data set, with the t0 already calibtrated

    Returns
    -------
    xarray.DataArray
        The updated xarray.DataArray
    """

    return assign_t0(da, da_ref.metadata.pump.t0_position)



import xarray as xr

from peaks.core.accessors.accessor_methods import register_lazy_accessor

# List of function names and the module they belong to
functions_to_register = {
    "fileIO.data_saving": ["save"],
    "display.plotting": ["plot_fit"],
    "fitting.fit": ["save_fit"],
}

# Register each function as a lazy accessor on Dataset
for module_name, func_names in functions_to_register.items():
    for func_name in func_names:
        accessor_property = register_lazy_accessor(
            func_name, f"peaks.core.{module_name}", xr.Dataset
        )



"""Functions that utilise machine learning techniques to denoise data."""

# Brendan Edwards 31/10/2023

import xarray as xr
from sklearn.decomposition import PCA

from peaks.core.accessors.accessor_methods import register_accessor


@register_accessor(xr.DataArray)
def SM_PCA(data, PCs=10):
    """Perform a principal component analysis on a spatial map, and then reconstruct the spatial map using the reduced
    dimensionality dataset and PCA eigenvectors. Results in a reduction in the information retained by the spatial map,
    but this lost information can just be noise. Thus, this function can act to markedly denoise spatial mapping data.
    The effect becomes more apparent for lower numbers of principal components, but too low will oversimplify data.

    Parameters
    ------------
    data : xarray.DataArray
        The spatial mapping data to perform a principal component analysis on.

    PCs : int, optional
        The number of principal components used to perform PCA. Defaults to 10.

    Returns
    ------------
    Reconstructed_SM : xarray.DataArray
        The reconstructed spatial map.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        SM = load('SM.ibw')

        # Use a 10 principal components PCA to reduce the information contained a spatial map, effectively denoising it
        reconstructed_SM1 = SM.SM_PCA()

        # Use a 30 principal components PCA to reduce the information contained a spatial map, effectively denoising it
        reconstructed_SM2 = SM.SM_PCA(PCs=30)

    """

    # Prevent unwanted overwriting of original data
    data = data.copy(deep=True)

    # Represent spatial mapping data as a tabular pandas DataFrame
    df = data.ML_pre_proc(extract="dispersion", scale=False)

    # Perform PCA
    pca = PCA(n_components=PCs)
    principal_components = pca.fit_transform(df)

    # Reconstruct the spatial map using the reduced dimensionality dataset and PCA eigenvectors
    reconstructed_data = pca.inverse_transform(principal_components).reshape(
        len(data.x1), len(data.x2), data.shape[2], data.shape[3]
    )

    # Represent reconstructed data using xarray format
    coords = (
        data.dims
    )  # Used so function works for any eV/k_par/theta_par combination/order
    reconstructed_SM = xr.DataArray(
        reconstructed_data,
        dims=("x1", "x2", coords[2], coords[3]),
        coords={
            "x1": data.x1,
            "x2": data.x2,
            coords[2]: data.coords[coords[2]],
            coords[3]: data.coords[coords[3]],
        },
    )
    reconstructed_SM.attrs = data.attrs

    return reconstructed_SM



import os
import pickle
import re

import pandas as pd
import xarray as xr


def _load_core_level_BEs():
    """Load core levels from the stored data."""
    path = os.path.join(os.path.dirname(__file__), "core_level_BEs.pkl")
    with open(path, "rb") as f:
        core_level_BEs = pickle.load(f)
    return core_level_BEs  # Return nested dictionary


def _load_core_level_xcs():
    """Load core level cross sections from the stored data."""
    # Load raw data
    root_path = os.path.dirname(__file__)
    with open(os.path.join(root_path, "xc_name.pkl"), "rb") as f:
        xc_name = pickle.load(f)
    with open(os.path.join(root_path, "xc.pkl"), "rb") as f:
        xc_data = pickle.load(f)

    # Sort the data in sensible order
    orbital_order = {"s": 0, "p": 1, "d": 2, "f": 3, "g": 4, "h": 5}

    def core_level_key(name):
        element, core_level = name.split("_")
        principal_quantum_number = int(
            core_level[0]
        )  # Extract the principal quantum number
        orbital = core_level[1]  # Extract the orbital character
        return (element, principal_quantum_number, orbital_order.get(orbital, 6))

    xc_name_sorted = sorted(xc_name, key=core_level_key)

    # Iterate through xc_name, xc_data, and En_entry to create DataArrays and store them
    # in a DataTree
    data_dict = {}
    for name in xc_name_sorted:
        try:
            cross_section = xc_data[f"xc_{name}"]
            energy = xc_data[f"En_{name}"]
            element, core_level = name.split("_")

            # Create a DataArray
            data_array = xr.DataArray(
                data=cross_section, dims=["eV"], coords={"eV": energy}, name="data"
            )

            # Add to a dictionary of Datasets
            dataset = xr.Dataset({"data": data_array})
            data_dict[f"{element}/{core_level}"] = dataset
        except KeyError:
            pass

    # Create and return DataTree from the dictionary
    return xr.DataTree.from_dict(data_dict)


class CoreLevels:
    """Helper tools for working with core levels in XPS.

    Class Attributes
    ----------------
    BE : dict
        Nested dictionary of core level binding energies.
    xc : xarray.DataTree
        DataTree of core level cross sections.

    Notes
    -----
    Binding energies are taken from J. A. Bearden and A. F. Burr, "Reevaluation of X-Ray
    Atomic Energy Levels," Rev. Mod. Phys. 39, (1967) p.125; M. Cardona and L. Ley, Eds.,
    Photoemission in Solids I: General Principles (Springer-Verlag, Berlin, 1978) with
    additional corrections, and from J. C. Fuggle and N. Mårtensson, "Core-Level Binding
    Energies in Metals," J. Electron Spectrosc. Relat. Phenom. 21, (1980) p.275.

    Cross sections from https://vuo.elettra.eu/services/elements/WebElements.html.
    """

    BE = _load_core_level_BEs()
    xc = _load_core_level_xcs()

    @staticmethod
    def _BE_to_KE(df, hv, order):
        """Convert data frame with core level data in BE to KE."""
        return (order * hv) - 4.35 - df

    @classmethod
    def _get_BE_df(cls, elements):
        """Get core level binding energies for specified elements as a dataframe."""
        if isinstance(elements, str):
            elements = [elements]
        BE_data = {element: cls.BE[element] for element in elements}
        return pd.DataFrame.from_dict(BE_data, orient="index")

    @classmethod
    def _get_KE_df(cls, df, hv, max_order):
        """Convert df of core level binding energies into kinetic energies."""
        rows = []
        for order in range(1, max_order + 1):
            new_idx = f"KE@hv={hv}eV"
            if order != 1:
                new_idx += (
                    f", {order}nd order"
                    if order == 2
                    else f", {order}rd order"
                    if order == 3
                    else f", {order}th order"
                )
            KE_df = cls._BE_to_KE(df, hv, order)
            KE_df.index = [f"{idx} ({new_idx})" for idx in KE_df.index]
            rows.append(KE_df)

        KE_df = pd.concat(rows)
        return KE_df.map(lambda x: x if x >= 0 else float("nan")).sort_index()

    @staticmethod
    def _clean_df(df):
        """Clean up a dataframe for display."""
        return df.round(2).fillna("").astype(str).sort_index()

    @classmethod
    def by_element(cls, elements, hv=None, max_order=1):
        """Get core level binding and optionally kinetic energies for specified elements.

        Parameters
        ----------
        elements : str | list
            Element symbol(s) to get core level binding energies for.
        hv : float, optional
            Photon energy in eV. If not specified, only binding energy returned.
        max_order : int, optional
            Maximum order of the light to consider. Default is 1.
        """
        df = cls._get_BE_df(elements)

        if hv is not None:
            KE_df = cls._get_KE_df(df, hv, max_order)
            df = pd.concat([df, KE_df])

        return cls._clean_df(df)

    @classmethod
    def by_energy(cls, energy, hv=None, max_order=1, tol=3):
        """Get core level binding and optionally kinetic energies for specified photon
        energy.

        Parameters
        ----------
        energy : float
            Core level energy in eV.
            If `hv=None`, search is performed on binding energy.
            If `hv` is specified, search is performed on kinetic energy from the
            specified photon energy, and optionally higher harmonics up to `max_order`.
        hv : float, optional
            Photon energy in eV.
        max_order : int, optional
            Maximum order of the light to consider. Default is 1.
        """

        df = pd.DataFrame.from_dict(cls.BE, orient="index")
        if hv is not None:
            df = cls._get_KE_df(df, hv, max_order)

        filtered_df = (
            df[(df >= energy - tol) & (df <= energy + tol)]
            .dropna(how="all")
            .dropna(axis=1, how="all")
        )

        return cls._clean_df(filtered_df)

    @classmethod
    def plot(
        cls,
        elements,
        eV=None,
        hv=None,
        max_order=1,
        ax=None,
        show_binding_as_negative=True,
        **kwargs,
    ):
        """Plot markers at core level binding or kinetic energies for specified elements.

        Parameters
        ----------
        elements : str | list
            Element symbol(s) to consider.
        eV: slice, optional
            Energy range in eV to plot.
        hv : float, optional
            Photon energy in eV, for plotting kinetic energy.
            If not specified, binding energy is plotted.
        max_order : int, optional
            Maximum order of the light to consider. Default is 1.
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If not specified, a new figure is created
        show_binding_as_negative : bool, optional
            If `True`, binding energies are shown as negative values to match the E-E_F
            convention for photoemission plotting.
        kwargs
            Additional keyword arguments passed to `matplotlib.pyplot.plot`.

        """

        import matplotlib.pyplot as plt

        if eV is None:
            eV = slice(None, None)

        def extract_element_and_order(s):
            # Define the regular expression pattern
            pattern = r"^(\w+)(?:\s+\(([^)]*order[^)]*)\))?"

            # Search for the pattern in the string
            match = re.search(pattern, s)

            if match:
                element = match.group(1)
                order = None
                if match.group(2):
                    order_match = re.search(r"(\d+\w* order)", match.group(2))
                    if order_match:
                        order = order_match.group(1)
                return element, order
            else:
                return None, None

        # Get the relevant data
        df = cls._get_BE_df(elements)
        if hv is not None:
            df = cls._get_KE_df(df, hv, max_order)
            ax_label = "Kinetic Energy (eV)"
        else:
            if show_binding_as_negative:
                df = -df
                ax_label = "E-E_F (eV)"
            else:
                ax_label = "Binding Energy (eV)"

        # Plot the data
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel(ax_label)

        # Set default plot options if not overwritten by kwargs
        color_cycle = ["k", "b", "g", "r", "c", "m", "y", "k"]
        color = kwargs.pop("color", None)
        linestyle = kwargs.pop("linestyle", None)

        elements = [extract_element_and_order(idx)[0] for idx in df.index]
        color_map = {
            element: color_cycle[i % len(color_cycle)]
            for i, element in enumerate(set(elements))
        }

        for idx, row in df.iterrows():
            element, order = extract_element_and_order(idx)
            row_vals = row.dropna()
            for cl, en in row_vals.items():
                # Plot core level if in range
                if en > (eV.start or -float("inf")) and en < (eV.stop or float("inf")):
                    ax.axvline(
                        en,
                        color=(color or color_map[element]),
                        linestyle=(linestyle or ("-" if order is None else "--")),
                        **kwargs,
                    )
                    # Label the core level
                    label = f"{element} {cl}"
                    if order:
                        label += f" ({order})"
                    ax.text(
                        en,
                        1.02,
                        label,
                        rotation=70,
                        verticalalignment="bottom",
                        horizontalalignment="left",
                        transform=plt.gca().get_xaxis_transform(),
                        color=(color or color_map[element]),
                    )

    @classmethod
    def _add_core_levels_to_ptab(cls, df, hv=None):
        # Add in core level binding energies to the df in the ptab function
        df["core levels"] = ""
        cl_df = cls._get_BE_df(cls.BE.keys())
        if hv is not None:
            cl_df = cls._get_KE_df(cl_df, hv, 1)
            cl_df.index = cl_df.index.str.extract(r"^(\w+)")[0]
        cl_df = cl_df.round(2)
        for element, element_df in cl_df.iterrows():
            element_cl_str = ""
            for level, value in element_df.dropna().items():
                element_cl_str += f"{level}: {value} eV<br>"
            df.loc[df["symbol"] == element, "core levels"] = element_cl_str

        return df

    @classmethod
    def ptab(cls, hv=None):
        """Interactive periodic table display showing core level energies.

        Parameters
        ----------
        hv : float, optional
            Photon energy in eV. If not specified, binding energy is displayed. If
            specified, kinetic energy is displayed.
        """
        from peaks.core.GUI.periodic_table.ptab import ptab

        if hv is None:
            tooltip = "Binding energy:<br>@{core levels}{safe}"
        else:
            tooltip = f"Kinetic energy (hv={hv} eV):<br>" + "@{core levels}{safe}"

        ptab(
            lambda df: cls._add_core_levels_to_ptab(df, hv=hv),
            ("Core levels", tooltip),
        )



# peaks/_lazy_import.py
"""_lzay_import.py

This module provides a class that allows for lazy importing of modules, i.e. to only import a module when it is called.

Classes:
- LazyImport: Class that allows for lazy importing of modules.

"""

import importlib

__all__ = ["LazyImport"]


class LazyImport:
    def __init__(self, name):
        self._name = name
        self._module = None

    def _load(self):
        if self._module is None:
            self._module = importlib.import_module(self._name)

    def __getattr__(self, attr):
        self._load()
        return getattr(self._module, attr)



"""Beamline/file-specific data loaders."""

# ruff: noqa: F401
from peaks.core.fileIO.loaders import (
    ape,
    bloch,
    casiopee,
    cif,
    clf,
    diamond,
    netcdf,
    standrews_mbs,
    standrews_phoibos,
    zarr,
)



import pyqtgraph as pg


class _Crosshair(pg.GraphicsObject):
    """Crosshair class for pyqtgraph plots.
    Creates a crosshair with a draggable centre marker and spans for integration ranges.
    """

    def __init__(
        self,
        plot,
        pos=(0, 0),
        dim0_width=0,
        dim1_width=0,
        *,
        orientation="both",
        bounds=None,
        movable=True,
        symbol="crosshair",
        brush=None,
        axisOrder="col-major",
        target_item_kwargs=None,
        linear_region_item_kwargs=None,
    ):
        """
        Initializes a Crosshair instance for use in 2D pyqtgraph plots, providing a moveable crosshair with variable
         spans. Defiend by dim0 and dim1 properties.

        Parameters
        ----------
        plot : PlotItem
            The plot item to which this crosshair will be added.
        pos : tuple of float, optional
            Initial position (dim0, dim1) of the crosshair's center. Defaults to (0, 0).
        dim0_width : float
            Width of the crosshair lines in dim0 direction. Defaults to 0.
        dim1_width : float
            Width of the crosshair lines in dim1 direction. Defaults to 0.
        orientation : str, optional
            Orientation of the crosshair. Options are "horizontal", "vertical", or "both". Defaults to "both".
        bounds : list of tuple of float, optional
            Lower and upper bounds for the crosshair. Provide list of tuple
            [(min_bound_dim0, max_bound_dim0), (min_bound_dim1, max_bound_dim1)].
            Defaults to None.
        movable : bool, optional
            If True, allows the crosshair to be moved interactively. Defaults to True.
        symbol : str, optional
            Symbol used for the crosshair center marker. Defaults to "crosshair".
        brush : QBrush, optional
            Brush used for drawing the crosshair spans. Defaults to None.
        axisOrder : str, optional
            Order of the axes. Defaults to "col-major" for consistency with ImageItem default.
        target_item_kwargs : dict, optional
            Additional keyword arguments to pass to the TargetItem constructor. Defaults to None.
        linear_region_item_kwargs : dict, optional
            Additional keyword arguments to pass to the LinearRegionItem constructor. Defaults to None.

        Raises
        ------
        ValueError
            If an unsupported orientation or axisOrder is specified.
        """

        super().__init__()
        self.parent_plot = plot
        self.orientation = orientation
        self.axis_order = axisOrder
        self.dim0_width = dim0_width
        self.dim1_width = dim1_width
        self.lock_x = None
        self.lock_y = None
        if target_item_kwargs is None:
            target_item_kwargs = {}
        if linear_region_item_kwargs is None:
            linear_region_item_kwargs = {}

        # Define horizontal and vertical based on axisOrder
        if self.axis_order == "col-major":  # --> (h, v) = (dim0, dim1)
            self.h_width = self.dim0_width
            self.v_width = self.dim1_width
        elif self.axis_order == "row-major":  # --> (h, v) = (dim1, dim0)
            pos = (pos[1], pos[0])
            self.h_width = self.dim1_width
            self.v_width = self.dim0_width
        else:
            raise ValueError(
                "Invalid axisOrder. Only 'col-major' or 'row-major' are allowed."
            )

        if bounds:
            self.set_bounds(bounds)

        # Crosshair items
        self.xh = pg.TargetItem(
            pos=pos,
            movable=movable,
            symbol=symbol,
            **target_item_kwargs,
        )  # Marker
        # Spans
        if self.orientation == "both" or self.orientation == "horizontal":
            self.xh_h = pg.LinearRegionItem(
                orientation="horizontal",
                values=(
                    pos[1] - self.v_width / 2,
                    pos[1] + self.v_width / 2,
                ),
                brush=brush,
                movable=False,
                **linear_region_item_kwargs,
            )
        if self.orientation == "both" or self.orientation == "vertical":
            self.xh_v = pg.LinearRegionItem(
                orientation="vertical",
                values=(
                    pos[0] - self.h_width / 2,
                    pos[0] + self.h_width / 2,
                ),
                brush=brush,
                movable=False,
                **linear_region_item_kwargs,
            )
        else:
            raise ValueError(
                "Invalid orientation. Only 'horizontal', 'vertical' or 'both' are allowed."
            )

        # Adding items to the parent
        # Check if self.xh_v and self.xh_h exist
        if hasattr(self, "xh_v"):
            self.parent_plot.addItem(self.xh_v)
        if hasattr(self, "xh_h"):
            self.parent_plot.addItem(self.xh_h)
        self.parent_plot.addItem(self.xh)

        # Connect signals
        self.xh.sigPositionChanged.connect(self.update_crosshair)

    def update_crosshair(self):
        """Update the crosshair position and span based on the marker position."""
        x, y = self.xh.pos()  # New position in actual plot coordinates
        # Ensure in range if bounds are set
        if hasattr(self, "h_bounds"):
            x = min(
                self.h_bounds[1] - self.h_width / 2,
                max(self.h_bounds[0] + self.h_width / 2, x),
            )
        if hasattr(self, "v_bounds"):
            y = min(
                self.v_bounds[1] - self.v_width / 2,
                max(self.v_bounds[0] + self.v_width / 2, y),
            )

        # Lock the crosshair to a specific value if required
        if self.lock_x:
            x = self.lock_x
        if self.lock_y:
            y = self.lock_y

        # Set the crosshair position
        self.xh.setPos((x, y))

        # Set span of LinearRegion items
        if hasattr(self, "xh_h"):
            self.xh_h.setRegion([y - self.v_width / 2, y + self.v_width / 2])
        if hasattr(self, "xh_v"):
            self.xh_v.setRegion([x - self.h_width / 2, x + self.h_width / 2])

    def set_dim0_width(self, width):
        """Set the dim0 width of the crosshair lines.

        Parameters
        ----------
        width : float
            Width of the span of the crosshair lines for the dim0 axis.
        """
        x, y = self.xh.pos()  # Current position in plot coordinates
        if self.axis_order == "col-major":
            self.h_width = width
            if hasattr(self, "xh_h"):
                self.xh_v.setRegion([x - self.h_width / 2, x + self.h_width / 2])
        else:
            self.v_width = width
            if hasattr(self, "xh_v"):
                self.xh_h.setRegion([y - self.v_width / 2, y + self.v_width / 2])

    def set_dim1_width(self, width):
        """Set the dim1 width of the crosshair lines.

        Parameters
        ----------
        width : float
            Width of the span of the crosshair lines for the dim1 axis.
        """
        x, y = self.xh.pos()  # Current position in plot coordinates
        if self.axis_order == "col-major":
            self.v_width = width
            if hasattr(self, "xh_v"):
                self.xh_h.setRegion([y - self.v_width / 2, y + self.v_width / 2])
        else:
            self.h_width = width
            if hasattr(self, "xh_h"):
                self.xh_v.setRegion([x - self.h_width / 2, x + self.h_width / 2])

    def get_dim0_width(self):
        """Get the dim0 width of the crosshair lines.

        Returns
        -------
        float
            Width of the span of the crosshair lines for the dim0 axis.
        """
        return self.h_width if self.axis_order == "col-major" else self.v_width

    def get_dim1_width(self):
        """Get the dim1 span of the crosshair lines.

        Returns
        -------
        float
            Width of the span of the crosshair lines for the dim1 axis.
        """
        return self.v_width if self.axis_order == "col-major" else self.h_width

    def get_dim0_span(self):
        """Get the dim0 span of the crosshair lines.

        Returns
        -------
        tuple of float
            Span of the crosshair lines for the dim0 axis.
        """
        if self.axis_order == "col-major":
            return self.xh_v.getRegion()
        else:
            return self.xh_h.getRegion()

    def get_dim1_span(self):
        """Get the dim1 span of the crosshair lines.

        Returns
        -------
        tuple of float
            Span of the crosshair lines for the dim1 axis.
        """
        if self.axis_order == "col-major":
            return self.xh_h.getRegion()
        else:
            return self.xh_v.getRegion()

    def set_pos(self, pos):
        """Set the position of the crosshair center.

        Parameters
        ----------
        pos : tuple of float
            Position (x, y) of the crosshair's center.
        """
        if self.axis_order == "row-major":
            pos = (pos[1], pos[0])

        self.xh.setPos(pos)
        if hasattr(self, "xh_h"):
            self.xh_h.setRegion([pos[1] - self.v_width / 2, pos[1] + self.v_width / 2])
        if hasattr(self, "xh_v"):
            self.xh_v.setRegion([pos[0] - self.h_width / 2, pos[0] + self.h_width / 2])

    def get_pos(self):
        """Get the position of the crosshair center in (dim0, dim1) order.

        Returns
        -------
        tuple of float
            Position of the crosshair's center in (dim0, dim1) order."""
        if self.axis_order == "col-major":
            x, y = self.xh.pos()
        else:
            y, x = self.xh.pos()
        return x, y

    def set_bounds(self, bounds):
        """Set the bounds of the crosshair.

        Parameters
        ----------
        bounds : list of tuple of float
            Lower and upper bounds for the crosshair. Provide list of tuple
            [(min_bound_dim0, max_bound_dim0), (min_bound_dim1, max_bound_dim1)].
        """
        if self.axis_order == "col-major":
            self.h_bounds = sorted(bounds[0])
            self.v_bounds = sorted(bounds[1])
        else:
            self.h_bounds = sorted(bounds[1])
            self.v_bounds = sorted(bounds[0])

    def set_lock_dim0(self, value):
        """Lock the crosshair to a specific dim0 value.

        Parameters
        ----------
        value : float or None
            Value to lock the crosshair to. Set with None to unlock.
        """
        if self.axis_order == "col-major":
            self.lock_x = value
        else:
            self.lock_y = value

    def set_lock_dim1(self, value):
        """Lock the crosshair to a specific dim1 value. Set with None to unlock."""
        if self.axis_order == "col-major":
            self.lock_y = value
        else:
            self.lock_x = value

    def set_visible(self, visible):
        """Set the crosshair visibility.

        Parameters
        ----------
        visible : bool
            True to show the crosshair, False to hide it.
        """
        self.xh.setVisible(visible)
        if hasattr(self, "xh_h"):
            self.xh_h.setVisible(visible)
        if hasattr(self, "xh_v"):
            self.xh_v.setVisible(visible)



"""Unsupervised clustering functions."""

# Brendan Edwards 31/10/2023

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from IPython.display import clear_output
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from tqdm import tqdm

from peaks.core.accessors.accessor_methods import register_accessor
from peaks.core.display.plotting import plot_grid
from peaks.utils import analysis_warning


@register_accessor(xr.DataArray)
def ML_pre_proc(
    data, extract="dispersion", E=0, dE=0, k=0, dk=0, scale=False, norm=False
):
    """Represent spatial mapping data as a tabular pandas DataFrame where each spatial position is a feature.
    This DataFrame format is used to represent data in machine learning functions.

    Parameters
    ------------
    data : xarray.DataArray
        The spatial mapping data to change to a tabular pandas dataframe.

    extract : str, optional
        Determines what is extracted from spatial mapping data. Defaults to dispersion. Valid entries are:
            dispersion
            MDC
            EDC
        Selecting MDC/EDC will rapidly increase calculation time.

    E : float, optional
        Energy of MDCs to extract. Defaults to 0.

    dE : float, optional
        MDC Integration range (represents the total range, i.e. integrates over +/- dE/2). Defaults to 0.

    k : float, optional
        k or theta_par value of EDCs to extract. Defaults to 0.

    dk : float, optional
        EDC integration range (represents the total range, i.e. integrates over +/- dk/2). Defaults to 0.

    scale : bool, optional
        Whether to apply standard scaling to data (i.e. center all values in each dimension around zero with unit
        standard deviation). Defaults to False.

    norm : bool, optional
        Whether to normalise the data at each spatial position. Defaults to False.

    Returns
    ------------
    df : pandas.DataFrame
        The spatial mapping data represented as a tabular pandas dataframe.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        SM = load('SM.ibw')

        # Get spatial mapping data in a tabular pandas dataframe format
        SM_df1 = SM.ML_pre_proc()

        # Get spatial mapping MDCs in a tabular pandas dataframe format
        SM_df2 = SM.ML_pre_proc(extract='MDC', E=73.42, dE=0.02)

    """

    # List to store spatial mapping data
    sample_data = []

    # Loop through spatial positions
    x1_vals = data.x1.data
    x2_vals = data.x2.data
    for i in range(len(x1_vals)):
        for j in range(len(x2_vals)):
            current_disp = data.isel(x1=i).isel(x2=j)

            # Apply normalisation at each spatial position if requested
            if norm:
                current_disp = current_disp.norm()

            # If we want to represent each feature to be a full dispersion
            if extract == "dispersion":
                # Get total number of dimensions (pixels) of dispersion
                dimensions = 1
                for coord_dim in current_disp.shape:
                    dimensions = dimensions * coord_dim

                # Reshape current dispersion into tabular form and append row to data
                reshaped_disp = current_disp.data.reshape(dimensions)
                sample_data.append(reshaped_disp)

            # If we want to represent each feature to be an MDC
            elif extract == "MDC":
                # Extract MDC from dispersion
                current_MDC = current_disp.MDC(E=E, dE=dE)
                sample_data.append(current_MDC.data)

            # If we want to represent each feature to be an EDC
            elif extract == "EDC":
                # Extract EDC from dispersion
                current_EDC = current_disp.EDC(k=k, dk=dk)
                sample_data.append(current_EDC.data)

            # Else user has entered an invalid method argument
            else:
                raise Exception("Method must be dispersion, MDC or EDC.")

    # Create tabular pandas dataframe
    df = pd.DataFrame(data=np.array(sample_data))

    if scale:
        # Apply optional standard scaling (center all values in each dimension around zero with unit standard deviation)
        std_scaler = StandardScaler()
        df = pd.DataFrame(data=std_scaler.fit_transform(df))

    return df


def perform_k_means(data, k=3, n_init="auto"):
    """Perform k-means clustering using scikit-learn.

    Parameters
    ------------
    data : pandas.DataFrame
        The data represented as a tabular pandas dataframe to perform clustering analysis on.

    k : int, optional
        Number of clusters. Defaults to 3.

    n_init : int, string, optional
        Number of times the k-means algorithm will be run with different centroid seeds. The final results will be the
        best output of n_init consecutive runs in terms of inertia. Required since the kmeans algorithm can fall into
        local minima, so repeats are required to check for this. Defaults to 'auto' (note: if an outdated scikit-learn
        package is installed where 'auto' is not yet implemented, an error will arise. In this case set n_init=10, or
        similar).

    Returns
    ------------
    model : sklearn.cluster._kmeans.KMeans
        K-means clustering analysis model information.

    labels : numpy.ndarray
        Array of spatially-dependent cluster assignments.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        SM = load('SM.ibw')

        # Get spatial mapping data in a tabular pandas dataframe format
        SM_df = SM.ML_pre_proc()

        # Perform k-means clustering analysis for 3 clusters
        model1, labels1 = perform_k_means(SM_df)

        # Perform k-means clustering analysis for 4 clusters
        model2, labels2 = perform_k_means(SM_df, k=4)

    """

    # Create a k-means clustering model
    model = KMeans(n_clusters=k, n_init=n_init)

    # Fit the model to the sampling data
    model.fit(data)

    # Predict the labels of the samples, i.e. which cluster they belong to
    labels = model.predict(data)

    return model, labels


@register_accessor(xr.DataArray)
def clusters_explore(
    data,
    cluster_range=None,
    n_init="auto",
    use_PCA=True,
    PCs=3,
    extract="dispersion",
    E=0,
    dE=0,
    k=0,
    dk=0,
    scale=False,
    norm=False,
):
    """Perform an exploratory k-means clustering analysis on a spatial map for a range of number of clusters.

    Parameters
    ------------
    data : xarray.DataArray
        The spatial mapping data to perform an exploratory k-means clustering analysis on.

    cluster_range : range, optional
        Range of number of clusters to perform k-means clustering analysis for. Defaults to range(1,7).

    n_init : int, string, optional
        Number of times the k-means algorithm will be run with different centroid seeds. The final results will be the
        best output of n_init consecutive runs in terms of inertia. Required since the kmeans algorithm can fall into
        local minima, so repeats are required to check for this. Defaults to 'auto' (note: if an outdated scikit-learn
        package is installed where 'auto' is not yet implemented, an error will arise. In this case set n_init=10, or
        similar).

    use_PCA : bool, optional
        Whether to apply a principal component analysis to the data. Defaults to True.

    PCs: int, optional
        Number of principal components. Defaults to 3.

    extract : str, optional
        Determines what is extracted from spatial mapping data. Defaults to dispersion. Valid entries are:
            dispersion
            MDC
            EDC
        Selecting MDC/EDC will rapidly increase calculation time.

    E : float, optional
        Energy of MDCs to extract. Defaults to 0.

    dE : float, optional
        MDC Integration range (represents the total range, i.e. integrates over +/- dE/2). Defaults to 0.

    k : float, optional
        k or theta_par value of EDCs to extract. Defaults to 0.

    dk : float, optional
        EDC integration range (represents the total range, i.e. integrates over +/- dk/2). Defaults to 0.

    scale : bool, optional
        Whether to apply standard scaling to data (i.e. center all values in each dimension around zero with unit
        standard deviation). Defaults to False.

    norm : bool, optional
        Whether to normalise the data at each spatial position. Defaults to False.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        SM = load('SM.ibw')

        # Perform an exploratory k-means clustering analysis for numbers of clusters ranging from 1 to 6, applying a
        # principal component analysis to the spatial mapping data using 3 principal components
        SM.clusters_explore()

        # Perform an exploratory k-means clustering analysis on spatial mapping MDCs for numbers of clusters ranging
        # from 1 to 10
        SM.clusters_explore(cluster_range=range(1,11), use_PCA=False, extract='MDC', E=73.42, dE=0.02)

    """
    if cluster_range is None:
        # Default range of number of clusters to test
        cluster_range = range(1, 7)

    # Prevent unwanted overwriting of original data
    data = data.copy(deep=True)

    # Represent spatial mapping data as a tabular pandas dataframe
    df = data.ML_pre_proc(
        extract=extract, E=E, dE=dE, k=k, dk=dk, scale=scale, norm=norm
    )

    # Perform principal component analysis if PCA is True
    if use_PCA:
        pca = PCA(n_components=PCs)  # Define PCA model
        principal_components = pca.fit_transform(
            df
        )  # Fit PCA model to data and get principal components
        df = pd.DataFrame(
            data=principal_components
        )  # Get principal components in tabular pandas dataframe format

    # Perform k-means clustering analysis for the range of number of clusters (k) requested
    k_titles = ["k=" + str(k) for k in cluster_range]  # Titles for plots
    inertias = []  # Empty list to store model inertias (a metric that defines spread of a cluster)
    classification_maps = []  # Empty list to store classification maps (spatial maps of cluster labels)
    for num_clusters in tqdm(cluster_range, desc="Calculating", colour="CYAN"):
        model, labels = perform_k_means(
            data=df, k=num_clusters, n_init=n_init
        )  # Perform k-means clustering
        inertias.append(model.inertia_)
        classification_map_data = labels.reshape(
            len(data.x1), len(data.x2)
        )  # Reshape 1D labels to 2D
        classification_map = xr.DataArray(
            classification_map_data,
            dims=("x1", "x2"),
            coords={"x1": data.x1, "x2": data.x2},
        )
        classification_maps.append(classification_map)

    inertias_xarray = xr.DataArray(
        inertias, dims="num_clusters", coords={"num_clusters": cluster_range}
    )  # Create xarray of inertias

    # Find optimal number of clusters
    for item in abs(inertias_xarray.differentiate("num_clusters")).norm():
        # If rate of decrease in inertia is below 20% of the initial decrease, found optimal number of clusters
        if float(item) < 0.2:
            recommended_num_clusters = int(item.num_clusters)
            break
        # Else continue through loop
        else:
            recommended_num_clusters = int(item.num_clusters)

    # Remove progress bar
    clear_output()

    # Plot model inertia against number of clusters, and indicate recommended number of clusters
    inertias_xarray.plot(marker="o", figsize=(15, 4))
    plt.xlabel("Number of clusters (k)")
    plt.ylabel("Inertia")
    plt.axvline(recommended_num_clusters, c="black", linestyle="--")
    plt.title(r"Optimal k$\approx$" + str(recommended_num_clusters))
    plt.show()

    # Plot classification map dependence on number of clusters
    plot_grid(classification_maps, ncols=3, titles=k_titles, cmap="cividis", y="x2")
    plt.show()


@register_accessor(xr.DataArray)
def clusters(
    data,
    num_clusters=3,
    n_init="auto",
    use_PCA=True,
    PCs=3,
    extract="dispersion",
    E=0,
    dE=0,
    k=0,
    dk=0,
    scale=False,
    norm=False,
    robust=False,
    vmin=None,
    vmax=None,
):
    """Perform a k-means clustering analysis on a spatial map.

    Parameters
    ------------
    data : xarray.DataArray
        The spatial mapping data to perform an exploratory k-means clustering analysis on.

    num_clusters : int, optional
        Number of clusters to perform k-means clustering analysis for. Defaults to 3.

    n_init : int, string, optional
        Number of times the k-means algorithm will be run with different centroid seeds. The final results will be the
        best output of n_init consecutive runs in terms of inertia. Required since the kmeans algorithm can fall into
        local minima, so repeats are required to check for this. Defaults to 'auto' (note: if an outdated scikit-learn
        package is installed where 'auto' is not yet implemented, an error will arise. In this case set n_init=10, or
        similar).

    use_PCA : bool, optional
        Whether to apply a principal component analysis to the data. Defaults to True.

    PCs: int, optional
        Number of principal components. Defaults to 3.

    extract : str, optional
        Determines what is extracted from spatial mapping data. Defaults to dispersion. Valid entries are:
            dispersion
            MDC
            EDC
        Selecting MDC/EDC will rapidly increase calculation time.

    E : float, optional
        Energy of MDCs to extract. Defaults to 0.

    dE : float, optional
        MDC Integration range (represents the total range, i.e. integrates over +/- dE/2). Defaults to 0.

    k : float, optional
        k or theta_par value of EDCs to extract. Defaults to 0.

    dk : float, optional
        EDC integration range (represents the total range, i.e. integrates over +/- dk/2). Defaults to 0.

    scale : bool, optional
        Whether to apply standard scaling to data (i.e. center all values in each dimension around zero with unit
        standard deviation). Defaults to False.

    norm : bool, optional
        Whether to normalise the data at each spatial position. Defaults to False.

    robust : bool, optional
        Whether the argument robust=True is passed to the plots. Defaults to False.

    vmin : float, optional
        Matplotlib vmin value used in plots of dispersions. Defaults to None.

    vmax : float, optional
        Matplotlib vmax value used in plots of dispersions. Defaults to None.

    Returns
    ------------
    classification_map : xarray.DataArray
        Spatial map of cluster labels.

    cluster_center_disps : list
        Dispersions for each cluster center.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        SM = load('SM.ibw')

        # Perform a k-means clustering analysis using 3 clusters, applying a principal component analysis to the spatial
        # mapping data using 4 principal components
        cluster_map1, cluster_disps1 = SM.clusters(num_clusters=3, PCs=4)

        # Perform a k-means clustering analysis using 3 clusters on spatial mapping MDCs
        cluster_map2, cluster_disps2 = SM.clusters(num_clusters=3, use_PCA=False, extract='MDC', E=73.42, dE=0.02)

    """

    # Prevent unwanted overwriting of original data
    data = data.copy(deep=True)

    # Represent spatial mapping data as a tabular pandas dataframe
    df = data.ML_pre_proc(
        extract=extract, E=E, dE=dE, k=k, dk=dk, scale=scale, norm=norm
    )

    # Perform principal component analysis if PCA is True
    if use_PCA:
        pca = PCA(n_components=PCs)  # Define PCA model
        principal_components = pca.fit_transform(
            df
        )  # Fit PCA model to data and get principal components
        df = pd.DataFrame(
            data=principal_components
        )  # Get principal components in tabular pandas dataframe format

    # Perform k-means clustering analysis for the number of clusters (k) requested
    model, labels = perform_k_means(data=df, k=num_clusters, n_init=n_init)
    classification_map_data = labels.reshape(
        len(data.x1), len(data.x2)
    )  # Reshape 1D labels to 2D
    classification_map = xr.DataArray(
        classification_map_data,
        dims=("x1", "x2"),
        coords={"x1": data.x1, "x2": data.x2},
    )

    # Extract cluster center of each cluster (not done using cluster centers so that we can get dispersions if MDC/EDC
    # extraction is used)
    cluster_average_disps = []  # Empty list to store average dispersions
    # Loop through cluster labels
    for cluster in range(num_clusters):
        current_disps = []
        # Loop through spatial positions
        for i in range(len(data.x1)):
            for j in range(len(data.x2)):
                # If current spatial position is assigned to current cluster
                if float(classification_map.isel(x1=i).isel(x2=j)) == cluster:
                    current_disps.append(data.isel(x1=i).isel(x2=j).copy(deep=True))
        # Sum dispersions with the same cluster label
        total_disp = current_disps[0]
        for i in range(1, len(current_disps)):
            total_disp += current_disps[i]
        cluster_average_disps.append(total_disp / len(current_disps))

    # Find maximum vmax of the averaged cluster dispersions so intensity variations between terminations can be observed
    if not robust and not vmax:
        max_cluster_disps_vmax = 0
        for disp in cluster_average_disps:
            if float(disp.max()) > max_cluster_disps_vmax:
                max_cluster_disps_vmax = float(disp.max())

    # Plot integrated spectral weight and k-means clustering labels spatial maps
    fig, axes = plt.subplots(figsize=(12, 5), ncols=2)
    data.tot().plot(
        ax=axes[0], cmap="cividis", y="x2", cbar_kwargs={"label": "Intensity"}
    )
    classification_map.plot(
        ax=axes[1],
        cmap="cividis",
        y="x2",
        cbar_kwargs={"ticks": range(num_clusters), "label": "Cluster"},
    )
    axes[0].set_title("Integrated spectral weight")
    axes[1].set_title("k-means clustering (k=" + str(num_clusters) + ")")
    plt.tight_layout()
    plt.show()

    # Plot average dispersion of each cluster
    cluster_average_titles = [
        "Cluster " + str(i) for i in range(num_clusters)
    ]  # Titles for plots
    if vmax:
        plot_grid(
            cluster_average_disps,
            titles=cluster_average_titles,
            cmap="binary",
            y="eV",
            ncols=num_clusters,
            vmin=vmin,
            vmax=vmax,
            cbar_kwargs={"label": None},
        )
    else:
        if not robust:
            plot_grid(
                cluster_average_disps,
                titles=cluster_average_titles,
                cmap="binary",
                y="eV",
                ncols=num_clusters,
                vmin=vmin,
                vmax=max_cluster_disps_vmax,
                cbar_kwargs={"label": None},
            )
        else:
            plot_grid(
                cluster_average_disps,
                titles=cluster_average_titles,
                cmap="binary",
                y="eV",
                ncols=num_clusters,
                robust=True,
                cbar_kwargs={"label": None},
            )
    plt.tight_layout()
    plt.show()

    # If a PCA was used on full dispersions, plot also the reconstructed cluster centers
    if use_PCA and extract == "dispersion":
        # Get non-spatial coordinates
        coords = list(data.dims)
        coords.remove("x1")
        coords.remove("x2")

        # Get reconstructed cluster center dispersions from reduced dimensionality dataset
        reconstructed_cluster_centers = pca.inverse_transform(
            model.cluster_centers_
        ).reshape(num_clusters, len(data.coords[coords[0]]), len(data.coords[coords[1]]))
        reconstructed_cluster_centers_disps = []
        for center in reconstructed_cluster_centers:
            disp_xarray = xr.DataArray(
                center,
                dims=(coords[0], coords[1]),
                coords={
                    coords[0]: data.coords[coords[0]],
                    coords[1]: data.coords[coords[1]],
                },
            )
            reconstructed_cluster_centers_disps.append(disp_xarray)

        # Find maximum vmax of the reconstructed cluster center dispersions so intensity variations between
        # terminations can be observed
        if not robust and not vmax:
            max_reconstructed_cluster_disps_vmax = 0
            for disp in reconstructed_cluster_centers_disps:
                if float(disp.max()) > max_reconstructed_cluster_disps_vmax:
                    max_reconstructed_cluster_disps_vmax = float(disp.max())

        # reconstructed cluster center dispersions
        reconstructed_cluster_centers_titles = [
            "Cluster " + str(i) + " (reconstructed)" for i in range(num_clusters)
        ]  # Titles for plots
        if vmax:
            plot_grid(
                reconstructed_cluster_centers_disps,
                titles=reconstructed_cluster_centers_titles,
                cmap="binary",
                y="eV",
                ncols=num_clusters,
                vmin=vmin,
                vmax=vmax,
                cbar_kwargs={"label": None},
            )
        else:
            if not robust:
                plot_grid(
                    reconstructed_cluster_centers_disps,
                    titles=reconstructed_cluster_centers_titles,
                    cmap="binary",
                    y="eV",
                    ncols=num_clusters,
                    vmin=vmin,
                    vmax=max_reconstructed_cluster_disps_vmax,
                    cbar_kwargs={"label": None},
                )
            else:
                plot_grid(
                    reconstructed_cluster_centers_disps,
                    titles=reconstructed_cluster_centers_titles,
                    cmap="binary",
                    y="eV",
                    ncols=num_clusters,
                    robust=True,
                    cbar_kwargs={"label": None},
                )
        plt.tight_layout()
        plt.show()

    return classification_map, cluster_average_disps


@register_accessor(xr.DataArray)
def PCA_explore(
    data,
    PCs_range=None,
    threshold=0.95,
    extract="dispersion",
    E=0,
    dE=0,
    k=0,
    dk=0,
    scale=False,
    norm=False,
):
    """Perform an exploratory principal component analysis on a spatial map for a range of principal components.

    Parameters
    ------------
    data : xarray.DataArray
        The spatial mapping data to perform an exploratory principal component analysis on.

    PCs_range : range, optional
        Range of number of principal components to perform PCA for. Defaults to range(1,6).

    threshold : float, optional
        Required threshold for the explained variance fraction of the dimensionally reduced dataset. Defaults to 0.95.

    extract : str, optional
        Determines what is extracted from spatial mapping data. Defaults to dispersion. Valid entries are:
            dispersion
            MDC
            EDC
        Selecting MDC/EDC will rapidly increase calculation time.

    E : float, optional
        Energy of MDCs to extract. Defaults to 0.

    dE : float, optional
        MDC Integration range (represents the total range, i.e. integrates over +/- dE/2). Defaults to 0.

    k : float, optional
        k or theta_par value of EDCs to extract. Defaults to 0.

    dk : float, optional
        EDC integration range (represents the total range, i.e. integrates over +/- dk/2). Defaults to 0.

    scale : bool, optional
        Whether to apply standard scaling to data (i.e. center all values in each dimension around zero with unit
        standard deviation). Defaults to False.

    norm : bool, optional
        Whether to normalise the data at each spatial position. Defaults to False.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        SM = load('SM.ibw')

        # Perform an exploratory PCA on spatial mapping data for numbers of principal components ranging from 1 to 5
        SM.PCA_explore()

        # Perform an exploratory PCA on spatial mapping MDCs for numbers of principal components ranging from 1 to 10
        SM.PCA_explore(PCs_range=range(1,11), extract='MDC', E=73.42, dE=0.02)

    """
    if PCs_range is None:
        # Default range of principal components to test
        PCs_range = range(1, 6)

    # Prevent unwanted overwriting of original data
    data = data.copy(deep=True)

    # Represent spatial mapping data as a tabular pandas dataframes
    df = data.ML_pre_proc(
        extract=extract, E=E, dE=dE, k=k, dk=dk, scale=scale, norm=norm
    )

    summed_var_ratio = []  # Empty list to store explained variance fractions

    # Loop through the range of principal components to test, and perform PCA at each
    for num_principal_components in tqdm(PCs_range, desc="Calculating", colour="CYAN"):
        model = PCA(n_components=num_principal_components)  # Define PCA model
        model.fit(df)  # Fit PCA model to dataset
        summed_var_ratio.append(
            np.sum(model.explained_variance_ratio_)
        )  # Append the total explained variance fractions of each model
    clear_output()  # Remove progress bar

    # Check minimum number of principal axes required to exceed explained variance fraction threshold
    num_principal_axes = "Unknown"
    for i, item in enumerate(summed_var_ratio):
        if item >= threshold:
            num_principal_axes = PCs_range[i]
            break

    # If explained variance could not exceed threshold, display a warning
    if num_principal_axes == "Unknown":
        analysis_warning(
            "Explained variance could not exceed threshold. Either reduce threshold or "
            "increase PCA_range.",
            title="Analysis info",
            warn_type="danger",
        )

    # Plot results of exploratory PCA, and suggest minimum required number of principal axes (if the explained
    # variance threshold has been exceeded)
    plt.figure(figsize=(8, 5))
    plt.plot(PCs_range, summed_var_ratio, "o-")
    plt.axhline(threshold, c="black", linestyle="--")
    if num_principal_axes != "Unknown":
        plt.axvline(num_principal_axes, c="black", linestyle="--")
    plt.xlabel("Principal axes")
    plt.ylabel("Explained variance fraction")
    plt.title(r"Minimum number of required principal axes$=$" + str(num_principal_axes))
    plt.show()



"""Registry for file loaders and locations."""

import os
from os.path import isfile, join

import h5py
import natsort

LOC_REGISTRY = {}


def locs():
    """Return the list of available locations."""
    return set(LOC_REGISTRY.copy().keys())


def register_loader(loader_class):
    """Decorator to register a loader class in the LOC_REGISTRY."""
    LOC_REGISTRY[loader_class._loc_name] = loader_class
    return loader_class


class IdentifyLoc:
    """Class to identify the location based on the file name and its contents.

    The methods in this class follow a specific naming convention to handle different file formats.
    Except for the _default and _no_extension cases, each method name should start with `_handler_`
    followed by the file extension it handles. For example, `_handler_txt` handles `.txt` files,
    `_handler_zip` handles `.zip` files, etc.
    """

    @staticmethod
    def _default_handler(fname):
        raise ValueError(
            f"Location of file {fname} could not be determined. Please specify the location in the call to `load`"
        )

    @staticmethod
    def _no_extension(fname):
        """Handle case for no extension (normally folder supplied)"""
        # If there is no extension, the data is in a folder. This is consistent with SOLEIL CASSIOPEE Fermi maps or
        # CLF Artemis data

        # Extract identifiable data from the file to determine if the location is SOLEIL CASSIOPEE
        file_list = natsort.natsorted(os.listdir(fname))
        file_list_ROI = [
            item for item in file_list if "ROI1_" in item and isfile(join(fname, item))
        ]
        if len(file_list_ROI) > 1:  # Must be SOLEIL CASSIOPEE
            return "Soleil_Casiopee_ARPES"
        else:  # Likely CLF Artemis
            # Extract identifiable data from the file to determine if the location is CLF Artemis
            file_list_Neq = [item for item in file_list if "N=" in item]
            if len(file_list_Neq) > 0:  # Must be CLF Artemis
                return "CLF_Artemis"
        # If we are unable to determine the location, fallback to the default handler
        IdentifyLoc._default_handler(fname)

    @staticmethod
    def _handler_xy(fname):
        # If the file is .xy format, the location must be either MAX IV Bloch-spin, StA-Phoibos or StA-Bruker

        # Open the file and load the first line
        with open(fname) as f:
            line0 = f.readline()

        # If measurement was performed using Specs analyser, location must be MAX IV Bloch-spin or StA-Phoibos
        if "SpecsLab" in line0:
            with open(fname) as f:
                for _i in range(30):
                    # If the 'PhoibosSpin' identifier is present in any of the lines, location must be MAX IV Bloch-spin
                    if "PhoibosSpin" in f.readline():
                        return "MAX IV Bloch-spin"
            # Otherwise by default, assume StA-Phoibos
            return "StA_Phoibos"

        IdentifyLoc._default_handler(fname)

    @staticmethod
    def _handler_sp2(fname):
        # If the file is .sp2 format, the location must be MAX IV Bloch-spin
        return "MAX IV Bloch-spin"

    @staticmethod
    def _handler_krx(fname):
        # If the file is .krx format, the location must be StA-MBS
        return "StA_MBS"

    @staticmethod
    def _handler_txt(fname):
        # If the file is .txt format, the location must be StA-MBS, MAX IV Bloch, Elettra APE or SOLEIL CASSIOPEE

        # Open the file and load the first line
        with open(fname) as f:
            line0 = f.readline()

        # MAX IV Bloch, Elettra APE or SOLEIL CASSIOPEE .txt files follow the same SES data format, so we can identify
        # the location from the location line in the file
        if line0 == "[Info]\n":  # Identifier of the SES data format
            from peaks.core.fileIO.base_data_classes.base_data_class import (
                BaseDataLoader,
            )

            # Extract metadata dictionary from file
            metadata = BaseDataLoader.load_metadata(fname, loc="SES", quiet=True)
            location_identifier = metadata.get("local_location_identifier")
            if (
                "bloch" in location_identifier.lower()
                or "maxiv" in location_identifier.lower()
            ):
                return "MAXIV_Bloch_A"
            elif (
                "ape" in location_identifier.lower()
                or "elettra" in location_identifier.lower()
            ):
                return "Elettra_APE_LE"
            elif (
                "cassiopee" in location_identifier.lower()
                or "soleil" in location_identifier.lower()
            ):
                return "Soleil_Casiopee_ARPES"

        elif "Lines" in line0:
            # This should be MBS format, default to StA loader
            return "StA_MBS"

    @staticmethod
    def _handler_zip(fname):
        # If the file is .zip format, the file must be of SES format. Thus, the location must be MAX IV Bloch,
        # Elettra APE, SOLEIL CASSIOPEE or Diamond I05-nano (defl map)
        from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader

        # Extract metadata dictionary from file
        metadata = BaseDataLoader.load_metadata(fname, loc="SES", quiet=True)
        location_identifier = metadata.get("local_location_identifier")
        if (
            "bloch" in location_identifier.lower()
            or "maxiv" in location_identifier.lower()
        ):
            return "MAXIV_Bloch_A"
        elif (
            "ape" in location_identifier.lower()
            or "elettra" in location_identifier.lower()
        ):
            return "Elettra_APE_LE"
        elif (
            "cassiopee" in location_identifier.lower()
            or "soleil" in location_identifier.lower()
        ):
            return "Soleil_Casiopee_ARPES"
        elif (
            "i05" in location_identifier.lower()
            or "diamond" in location_identifier.lower()
        ):
            return "Diamond I05-nano"

    @staticmethod
    def _handler_ibw(fname):
        from peaks.core.fileIO.base_data_classes.base_ibw_class import BaseIBWDataLoader

        # Read the wavenote
        wavenote = BaseIBWDataLoader._load_metadata(fname)["wavenote"]

        # If the file is .ibw format, the file is likely SES format.
        if "SES" in wavenote:
            from peaks.core.fileIO.base_arpes_data_classes.base_ses_class import (
                SESDataLoader,
            )

            metadata_dict_SES_keys = SESDataLoader._SES_metadata_to_dict_w_SES_keys(
                wavenote.split("\r")
            )
            location = metadata_dict_SES_keys.get("Location")
            if "bloch" in location.lower() or "maxiv" in location.lower():
                return "MAXIV_Bloch_A"
            elif "ape" in location.lower() or "elettra" in location.lower():
                return "Elettra_APE_LE"
            elif "cassiopee" in location.lower() or "soleil" in location.lower():
                return "Soleil_Casiopee_ARPES"
            else:  # Return general SES loader
                return "SES"
        # If we are unable to find a location, define location as a generic ibw file
        return "ibw"

    @staticmethod
    def _handler_nxs(fname):
        # If the file is .nxs format, the location should be Diamond I05-nano or Diamond I05-HR

        from peaks.core.fileIO.base_data_classes.base_hdf5_class import (
            BaseHDF5DataLoader,
        )

        # Open the file (read only)
        with h5py.File(fname, "r") as f:
            # .nxs files at Diamond and Alba contain approximately the same identifier format
            identifier = BaseHDF5DataLoader._extract_hdf5_value(
                f, "entry1/instrument/name"
            )
        # From the identifier, determine the location
        if "i05-1" in identifier:
            return "Diamond_I05_Nano-ARPES"
        elif "i05" in identifier:
            return "Diamond_I05_ARPES"

    @staticmethod
    def _handler_h5(fname):
        # Use CLF_Artemis loader which will in turn call the FeSuMa loader for the key things
        return "CLF_Artemis"

    @staticmethod
    def _handler_nc(fname):
        # If the file is .nc format, the location should be NetCDF
        return "NetCDF"

    @staticmethod
    def _handler_zarr(fname):
        # If the file is .zarr format, the location should be Zarr
        return "Zarr"

    @staticmethod
    def _handler_cif(fname):
        # If the file is .cif format, the location should be cif
        return "cif"



import inspect
import os
from datetime import datetime

import pint_xarray  # noqa: F401
import xarray as xr

from peaks.core.fileIO.loc_registry import LOC_REGISTRY, IdentifyLoc
from peaks.core.metadata.base_metadata_models import (
    BaseScanMetadataModel,
)
from peaks.core.utils.misc import analysis_warning

ureg = pint_xarray.unit_registry


class BaseDataLoader:
    """Base Class for data loaders

    Notes
    -----
    At a minimum, subclasses should implement the `_load_data` and `_load_metadata` methods. They are also expected
    to define the `_loc_name` class variable, which is a string identifier for the location at which the data was
    obtained. This is used to determine which loader to use when loading data.

    Storing of metadata is based aorund the use of Pydantic models and then stored in the attrs of the DataArray.
    The metadata models should generally be taken or derived from the models in the base_metadata_models.py file.
    The base class provides a timestamp from the last modification time of the file. If possible, this can be updated
    via a more robust method in the implemented `_load_metadata` method, returning a timestamp in the metadata
    dictionary with key `timestamp`. In genreal, subclasses should also define specific metadata parsers to map the
    metadata dictionary returned from `_load_metatdata` to the relevant Pydantic models and to apply these to the
    DataArray. Any subclass should define the `_metadata_parsers` class variable as a list of these methods to be
    called.

    See Also
    --------
    BaseDataLoader._apply_specific_metadata
    base_metadata_models.py
    """

    # Define core attributes as class variables
    _loc_name = "Base"  # Identifier for the location/loader
    _loc_description = "Base data loader class - generally not used directly"  # Description of the location
    _loc_url = None  # Link to some descriptor of the location
    _desired_dim_order = []  # List of the desired dimension order in final da
    _dtype = None  # Desired dtype for the main data
    _dorder = None  # Desired array order for the main data
    _metadata_cache = {}  # Cache for metadata
    _metadata_parsers = []  # List of metadata parsers to apply

    # Properties to access class variables
    @property
    def loc_name(self):
        """Return the location name."""
        return self._loc_name  # Identifier

    @property
    def loc_description(self):
        """Return the location description."""
        return self._loc_description

    @property
    def loc_url(self):
        """Return the location URL."""
        return self._loc_url

    @property
    def desired_dim_order(self):
        """Return the desired dimension order for the data."""
        return self._desired_dim_order

    @property
    def dtype(self):
        """Return the desired dtype for the main data."""
        return self._dtype  # dtype to force for main data

    @property
    def dorder(self):
        """Return the desired array order for the main data."""
        return self._dorder  # Order to force for main data

    @property
    def metadata_key_mappings(self):
        """Return the metadata key mappings."""
        return self._metadata_key_mappings

    @property
    def metadata_warn_if_missing(self):
        """Return the metadata key mappings."""
        return self._metadata_warn_if_missing

    @property
    def metadata_cache(self):
        """Return the metadata cache."""
        return self._metadata_cache

    # Public methods
    @classmethod
    def load(cls, fpath, lazy=None, loc=None, metadata=True, quiet=False, **kwargs):
        """Top-level method to load data and return a DataArray.

        Parameters
        ------------
        fpath : str, list
            Full file path of file to load.

        lazy : str, bool, optional
            Whether to load data in a lazily evaluated dask format. Set explicitly using True/False Boolean.
            Defaults to `None` where a file is only loaded in the dask format if its spectrum is above threshold
            set in `opts.FileIO.lazy_size`

        loc : str, optional
            Location identifier for where data was acquired. Defaults to `None`, where the location will be attempted
            to be automatically determined.

        metadata : bool, optional
            Whether to attempt to load metadata from the file. Defaults to True.

        quiet : bool, optional
            Whether to suppress analysis warnings when loading data. Defaults to False.

        kwargs : dict
            Additional keyword arguments to pass to the individual loaders.

        Returns
        ------------
        da : xarray.DataArray
            The loaded data as an xarray DataArray.

        Examples
        ------------
        Normally used via the `pks.load` core function. If needed to be accessed directly, example usage is as follows::

            from peaks.core.fileIO.base_data_classes import BaseDataLoader

            fpath = 'C:/User/Documents/Research/disp1.xy'  # Path to the file to be loaded
            da = BaseDataLoader.load(fpath)  # Load the data

        Notes
        -----
        This method will generally be run from a different class than the loader class for the specific file.
        If building a data loader by subclassing this, make sure so put any loc-specific logic in the
        _load method.
        """

        # Make sure the metadata cache for this file is empty
        cls._metadata_cache.pop(fpath, None)

        # Parse the loc
        loc = loc if loc else cls._get_loc(fpath)
        cls._check_valid_loc(loc)  # Check a valid loc
        # Trigger the loader for the correct loc
        loader_class = cls.get_loader(loc)
        return loader_class._load(fpath, lazy, metadata, quiet, **kwargs)

    @classmethod
    def load_metadata(
        cls,
        fpath,
        loc=None,
        return_as_dict=False,
        quiet=True,
        load_metadata_from_file=True,
    ):
        """Top-level method to load metadata and return it in a dictionary.

        Parameters
        ------------
        fpath : str
            Path to the file to be loaded.

        loc : str, optional
            The location at which the data was obtained. Defaults to `None` where the location will be attempted to be
            determined automatically.

        return_as_dict : bool, optional
            Whether to return the metadata as a simple dictionary of metadata_keys: values or as a dictionary
            of the parsed (i.e. structured) metadata models. Defaults to False.

        quiet : bool, optional
            Whether to suppress missing metadata warnings when loading data. Defaults to True.

        load_metadata_from_file : bool, optional
            Whether to load the actual metadata from the file. Defaults to True. If False, only the basic file
            metadata will be added, along with the empty metadata structure


        Returns
        ------------
        metadata_dict : dict
            The loaded metadata as a dictionary simple dictionary or a dictionary of parsed metadata models.

        Examples
        ------------
        Example usage is as follows::

            from peaks.core.fileIO.base_data_classes import BaseDataLoader

            fpath = 'C:/User/Documents/Research/disp1.xy'
            metadata = BaseDataLoader.load_metadata(fpath)  # Load the metadata

        Notes
        -----
        This will generally be run from a different class than the loader class for the specific file.
        If building a data loader by subclassing this, make sure so put any loc-specific logic in the
        _load_metadata method.
        """

        # Parse the loc - if the base class is used, determine the loc automatically and route to the right subclass
        if cls._loc_name == "Base":
            loc = loc if loc else cls._get_loc(fpath)
            cls._check_valid_loc(loc)  # Check a valid loc
            if loc != "Base":
                return cls.get_loader(loc).load_metadata(fpath, return_as_dict, quiet)
        # Otherwise, use the loc defined in the subclass
        loc = cls._loc_name

        # Parse some baseline metadata from the file
        # Extract a timestamp from last modification time - overwrite in subclass if more robust method available
        timestamp = os.path.getmtime(fpath)
        # Convert the timestamp to a human-readable format
        readable_timestamp = datetime.fromtimestamp(timestamp).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        metadata_dict = {"timestamp": readable_timestamp, "fpath": fpath}

        if load_metadata_from_file:
            # Method to extract any specific metadata, which should be updated in subclasses
            metadata_dict.update(
                {
                    k: v
                    for k, v in cls.get_loader(loc)._load_metadata(fpath).items()
                    if v is not None
                }
            )
        if return_as_dict:
            return metadata_dict
        parsed_metadata = {}
        parsed_metadata.update(cls._parse_general_metadata(metadata_dict))
        parsed_metadata.update(cls._parse_specific_metadata(metadata_dict, quiet))
        return parsed_metadata

    @staticmethod
    def get_loader(loc):
        """Get the loader class for the given location.

        Parameters
        ------------
        loc : str
            The location at which the data was obtained.

        Returns
        ------------
        loader_class : class
            The loader class for the given location.

        Raises
        ------------
        ValueError
            If no loader is found for the given location.

        Examples
        ------------
        Example usage is as follows::

                from peaks.core.fileIO.base_data_classes import BaseDataLoader

                loc = 'I05'

                # Get the loader class for the given location
                loader_class = BaseDataLoader.get_loader(loc)
        """
        loader_class = LOC_REGISTRY.get(loc)
        if loader_class:
            return loader_class
        else:
            raise ValueError(
                f"No loader found for location {loc}. Expected one of {set(LOC_REGISTRY.keys())}."
            )

    # Private methods
    @staticmethod
    def _get_loc(fpath):
        """This function determines the location at which the data was obtained.

        Parameters
        ------------
        fpath : str
            Path to the file to be loaded.

        Returns
        ------------
        loc : str
            The name of the location (typically a beamline).
        """
        file_extension = os.path.splitext(fpath)[1]
        # Define the handlers for the different file extensions
        # No extension
        handlers = {
            "": IdentifyLoc._no_extension,
        }
        # For ones with file extensions, generate the handlers from the methods in IdentifyLocation
        handlers.update(
            {
                f".{method_name.split('_handler_')[1]}": method
                for method_name, method in inspect.getmembers(
                    IdentifyLoc, predicate=inspect.isfunction
                )
                if method_name.startswith("_handler")
            }
        )
        handler = handlers.get(file_extension, IdentifyLoc._default_handler)

        return handler(fpath)

    @staticmethod
    def _check_valid_loc(loc):
        """Check if the location is valid."""
        if loc not in LOC_REGISTRY.keys():  # Check a valid loc
            raise ValueError(
                f"No loader defined for location {loc}. Specify one of {LOC_REGISTRY.keys()} or leave"
                f" loc argument empty to attempt to determine the location automatically."
            )

    @classmethod
    def _load(cls, fpath, lazy, metadata, quiet, **kwargs):
        """Generic method for loading the data."""
        # Load the actual data from the file
        data = cls._load_data(fpath, lazy, **kwargs)
        da = cls._make_dataarray(data)  # Convert to DataArray
        # Add a name to the DataArray
        da.name = fpath.split("/")[-1].split(".")[0]
        parsed_metadata = cls.load_metadata(
            fpath,
            loc=cls._loc_name,
            quiet=quiet or not metadata,
            load_metadata_from_file=metadata,
        )
        da.attrs.update(parsed_metadata)
        cls._metadata_cache.pop(fpath, None)  # Clear any metadata cache for this file
        # Apply any specific conventions and add a history of the load
        da = cls._apply_conventions(da)
        cls._add_load_history(da, fpath)
        return da

    @classmethod
    def _load_data(cls, fpath, lazy, **kwargs):
        """Load the data. To be implemented by subclasses.

        Parameters
        ------------
        fpath : str
            Path to the file to be loaded.

        lazy : bool or str
            Whether to load data in a lazily evaluated dask format. Set explicitly using True/False Boolean.

        kwargs : dict
            Additional keyword arguments to pass to the individual loaders.

        Notes
        -----
        The data should be returned as a dictionary containing the following keys:
            spectrum : the data to be converted
            dims : a list of dims in order corresponding to the data axis order
            coords : associated coords
            units : dict containing units for the data and dimensions

        If metadata needs to be loaded
        """
        raise NotImplementedError("Subclasses should implement this method.")

    @classmethod
    def _load_metadata(cls, fpath):
        """Load the metadata. Should return a dictionary `metadata_dict` mapping relevant metadata keys to values.
        Loaders that subclass this class should implement this method.

        Parameters
        ------------
        fpath : str
            Path to the file to be loaded.

        Returns
        ------------
        metadata_dict : dict
            Dictionary mapping metadata keys to values.
            Keys should be `peaks` notation, generally of the form subclass_item (e.g. `analyser_eV`)
            Values should be :class:`pint.Quantity` objects where possible.
        """
        raise NotImplementedError("Subclasses should implement this method.")

    @classmethod
    def _parse_general_metadata(cls, metadata_dict):
        """Apply general metadata to the DataArray - implemented irrespective of the loader."""

        fpath = metadata_dict.get("fpath")
        general_scan_metadata = BaseScanMetadataModel(
            name=fpath.split("/")[-1].split(".")[0],
            filepath=fpath,
            loc=cls._loc_name,
            timestamp=metadata_dict.get("timestamp"),
            scan_command=metadata_dict.get("scan_command"),
        )
        return {"_scan": general_scan_metadata}

    @classmethod
    def _parse_specific_metadata(cls, metadata_dict, quiet):
        """Method to orchastrate applying loader-specific metadata to the DataArray.

        Parameters
        ------------
        metadata_dict : dict
            Dictionary containing all available metadata as returned by `_load_metadata`.

        quiet : bool
            Whether to suppress missing metadata warnings when loading data.

        Returns
        ------------
        parsed_metadata : dict
            Dictionary containing all parsed metadata as Pydantic models.

        Notes
        -----
        Subclasses should implement a method `_parse_xxxxxx_metadata` to map metadata from the general metadata
        dictionary to the specific pydantic metadata model and apply this the the dataarray. The class attribute
        _metadata_parsers should provide a list of all of these methods to be called.

        These methods should return a tuple of ({'key': MetadataModel()}, list) where 'key' is the key that should
        be used in the :class:xarray.DataArray attributes for holding the relevant metadata, MetadataModel is the
        model as generated from `base_metadata_models.py` and list is a list of metadata keys that should be checked
        if they exist and a warning be raised if they are missing.
        """
        metadata_warning_list = []
        parsed_metadata = {}
        for parser in cls._metadata_parsers:
            _parser = getattr(cls, parser, None)
            if _parser is None:
                raise NotImplementedError(
                    f"Method {parser} not found in subclass {cls.__name__} or its parent classes. Implement this or "
                    f"change the entries of class attribute `_metadata_parsers`."
                )
            metadata_to_apply, metadata_to_add_to_warning_list = _parser(metadata_dict)
            parsed_metadata.update(metadata_to_apply)
            if metadata_to_add_to_warning_list:
                metadata_warning_list.extend(metadata_to_add_to_warning_list)
        if not quiet:
            cls._warn_metadata(metadata_dict, metadata_warning_list)
        return parsed_metadata

    @classmethod
    def _warn_metadata(cls, metadata_dict, metadata_warning_list=None):
        """Warn if any of the metadata fields in missing_metadata_warning_list are missing."""
        missing_metadata_to_warn = [
            item for item in metadata_warning_list if metadata_dict.get(item) is None
        ]
        if missing_metadata_to_warn:
            analysis_warning(
                f"Unable to extract metadata for: {missing_metadata_to_warn}. If you expected this to be in the "
                f"available metadata, update the {cls.__name__} metadata loader in {cls.__module__} to account for new "
                "file format.",
                title="Loading info",
                warn_type="warning",
            )

    @classmethod
    def _make_dataarray(cls, data):
        """Convert data into an xarray DataArray.

        Parameters
        ------------
        data : dict
            Dictionary containing:
                spectrum : the data to be converted
                dims : a list of dims in order corresponding to the data axis order
                coords : associated coords
                units : dict containing units for the data and dimensions
        """
        da = xr.DataArray(
            data["spectrum"],
            dims=data.get("dims", None),
            coords=data.get("coords", None),
            name="spectrum",  # Temporarily name as spectrum to allow unit quantification
        )
        return da.pint.quantify(data.get("units"))

    @classmethod
    def _apply_conventions(cls, da):
        """Apply relevant conventions to the DataArray."""

        # Ensure that all the DataArray dimension coordinates are ordered low to high
        for dim in da.dims:
            if len(da[dim]) > 1:
                if da[dim].data[-1] - da[dim].data[0] < 0:
                    da = da.reindex({dim: da[dim][::-1]})
        # Reorder dimensions if needed
        if cls.desired_dim_order:
            da = da.transpose(*cls._desired_dim_order, ..., missing_dims="ignore")
        # Set the appropriate data type and order
        if cls.dtype:
            da = da.astype(cls._dtype, order=cls._dorder)
        return da

    @classmethod
    def _add_load_history(cls, da, fpath):
        """Add a history of the load to the DataArray."""
        da.history.add(
            {
                "record": "Data loaded",
                "loc": cls._loc_name,
                "loader": cls.__name__,
                "file_name": fpath,
            }
        )



import pint_xarray

from peaks.core.fileIO.base_arpes_data_classes.base_mbs_class import MBSDataLoader
from peaks.core.fileIO.loc_registry import register_loader

ureg = pint_xarray.unit_registry


@register_loader
class StAndrewsMBS(MBSDataLoader):
    """Data loader for St Andrews MBS ARPES system, measured using A1 Soft."""

    _loc_name = "StA_MBS"
    _loc_description = "MBS A1 in King Group at St Andrews"
    _loc_url = "https://www.quantummatter.co.uk/king"
    _analyser_slit_angle = 90 * ureg.deg
    _manipulator_name_conventions = {
        "polar": "sapolar",
        "tilt": "satilt",
        "azi": "saazi",
        "x1": "say",
        "x2": "saz",
        "x3": "sax",
    }
    _temperature_exclude_from_metadata_warn = [
        "sample",
        "cryostat",
        "shield",
        "setpoint",
    ]
    _photon_exclude_from_metadata_warn = ["hv", "polarisation", "exit_slit"]
    _manipulator_exclude_from_metadata_warn = [
        axis for axis in _manipulator_name_conventions.keys()
    ]

    _manipulator_sign_conventions = {}

    _analyser_sign_conventions = {
        "theta_par": -1,
        "deflector_parallel": -1,  # consistent with theta_par
    }

    _MBS_metadata_key_mappings = {}
    _MBS_metadata_units = {}



from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader
from peaks.core.metadata.base_metadata_models import TemperatureMetadataModel


class BaseTemperatureDataLoader(BaseDataLoader):
    """Base class for data loaders for systems with sample temperature control.

    Subclasses should define the `_load_temperature_metadata` method to return a dictionary of relevant metadata
    values with keys of the form `temperature_item` where `item` is the names in the `_temperature_attributes` list,
    i.e. is given in :class:`peaks` convention. This method should return values as :class:`pint.Quantity` objects
    where possible to ensure units are appropriately captured and propagated. Alternatively, the main `_load_metadata`
    method can be overwritten to return the full metadata dictionary, including manipulator metadata.

    Subclasses should add any additional temperature attributes via the `_add_temperature_attributes` class variable,
     providing a list of additional attributes.

    See Also
    --------
    BaseDataLoader
    BaseDataLoader._load_metadata
    """

    # Define class variables
    _loc_name = "Default Temperature"
    _temperature_attributes = ["sample", "cryostat", "shield", "setpoint"]
    _temperature_exclude_from_metadata_warn = [
        "cryostat",
        "shield",
        "setpoint",
    ]  # List of attributes to ignore for metadata warnings
    _metadata_parsers = [
        "_parse_temperature_metadata"
    ]  # Specific metadata parsers to apply

    # Properties to access class variables
    @property
    def temperature_attributes(self):
        """Return the temperature attributes."""
        return self._temperature_attributes

    @classmethod
    def _parse_temperature_metadata(cls, metadata_dict):
        """Parse metadata specific to the temperature data."""

        # Build and populate the temperature metadata model
        temperature_metadata = TemperatureMetadataModel(
            sample=metadata_dict.get("temperature_sample"),
            cryostat=metadata_dict.get("temperature_cryostat"),
            shield=metadata_dict.get("temperature_shield"),
            setpoint=metadata_dict.get("temperature_setpoint"),
        )

        metadata_to_warn_if_missing = (
            f"temperature_{attribute}"
            for attribute in cls._temperature_attributes
            if attribute not in cls._temperature_exclude_from_metadata_warn
        )

        # Return the model, and a list of any metadata that should be warned if missing
        return {"_temperature": temperature_metadata}, metadata_to_warn_if_missing



from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader
from peaks.core.metadata.base_metadata_models import (
    PhotonMetadataModel,
    PumpPhotonMetadataModel,
)


class BasePhotonSourceDataLoader(BaseDataLoader):
    """Base class for data loaders for experiments involving photon sources.

    Subclasses should define the `_load_photon_metadata` method to return a dictionary of relevant metadata
    values with keys of the form `photon_item` where `item` is the names in the `_photon_attributes` list,
    i.e. is given in :class:`peaks` convention. This method should return values as :class:`pint.Quantity` objects
    where possible to ensure units are appropriately captured and propagated. Alternatively, the main `_load_metadata`
    method can be overwritten to return the full metadata dictionary, including manipulator metadata.

    Subclasses should add any additional photon attributes via the `_add_photon_attributes` class variable,
     providing a list of additional attributes.

    See Also
    --------
    BaseDataLoader
    BaseDataLoader._load_metadata
    """

    _loc_name = "Default Photon Source"
    _photon_attributes = ["hv", "polarisation", "exit_slit"]
    _photon_exclude_from_metadata_warn = [
        "polarisation",
        "exit_slit",
    ]  # List of attributes to ignore for metadata warnings
    _metadata_parsers = ["_parse_photon_metadata"]  # Specific metadata parsers to apply

    # Properties to access class variables
    @property
    def photon_attributes(self):
        """Return the photon attributes."""
        return self._photon_attributes

    @classmethod
    def _parse_photon_metadata(cls, metadata_dict):
        """Parse metadata specific to the photon data."""

        # Build and populate the photon metadata model
        photon_metadata = PhotonMetadataModel(
            hv=metadata_dict.get("photon_hv"),
            polarisation=metadata_dict.get("photon_polarisation"),
            exit_slit=metadata_dict.get("photon_exit_slit"),
        )

        metadata_to_warn_if_missing = (
            f"photon_{attribute}"
            for attribute in cls._photon_attributes
            if attribute not in cls._photon_exclude_from_metadata_warn
        )

        # Return the model, and a list of any metadata that should be warned if missing
        return {"_photon": photon_metadata}, metadata_to_warn_if_missing


class BasePumpProbeClass(BasePhotonSourceDataLoader):
    _pump_photon_attributes = ["hv", "polarisation", "delay", "power"]
    _pump_photon_exclude_from_metadata_warn = _pump_photon_attributes
    _metadata_parsers = ["_parse_pump_photon_metadata"]

    @property
    def pump_photon_attributes(self):
        return self._pump_photon_attributes

    @classmethod
    def _parse_pump_photon_metadata(cls, metadata_dict):
        pump_photon_metadata = PumpPhotonMetadataModel(
            hv=metadata_dict.get("pump_hv"),
            polarisation=metadata_dict.get("pump_polarisation"),
            power=metadata_dict.get("pump_power"),
            delay=metadata_dict.get("pump_delay"),
            t0_position=metadata_dict.get("pump_t0_position"),
        )

        metadata_to_warn_if_missing = (
            f"pump_{attribute}"
            for attribute in cls._pump_photon_attributes
            if attribute not in cls._pump_photon_exclude_from_metadata_warn
        )

        return {"_pump": pump_photon_metadata}, metadata_to_warn_if_missing



"""Functions used to apply Fermi level corrections to data."""

import numexpr as ne
import numpy as np
import xarray as xr

from peaks.core.utils.interpolation import _fast_bilinear_interpolate_rectilinear
from peaks.core.utils.misc import analysis_warning


def _get_EF_at_theta_par0(da):
    """Gets the kinetic energy of the DataArray corresponding to the Fermi level at theta_par=0.
    The `EF_correction` attribute will be taken if set, or estimated if not.

    Parameters
    ----------
    da : xarray.DataArray
        Data to use.

    Returns
    -------
    float or np.ndarray
        Kinetic energy of the Fermi level at the given theta_par value(s).
    """

    # Get EF_correction, estimating if not present
    if "hv" not in da.dims:
        if da.metadata.calibration.EF_correction is None:
            _add_estimated_EF(da)
        EF_fn = da.metadata.get_EF_correction()
        if isinstance(EF_fn, dict):
            return EF_fn["c0"]
        else:
            return EF_fn
    else:  # For photon energy scan, we need this from the coords
        if "EF" not in da.coords:
            _add_estimated_EF(da)
        return da.coords["EF"].data  # EF vs hv


def _get_E_shift_at_theta_par(da, theta_par, Ek=None):
    """Gets the energy shift that should be applied to a kinetic energy to correct for the curvature of the Fermi
    level as a function of theta_par. The `EF_correction` attribute should be set first.
    Uses numexpr to allow for fast evaluation even over large theta_par grids.

    Parameters
    ----------
    da : xarray.DataArray
        Underlying data
    theta_par : float or np.ndarray
        Theta_par value to extract the shift at.
    Ek : float or np.ndarray, optional
        Kinetic energy value(s) to apply the shift to. If not provided, the raw shift will be returned.

    Returns
    -------
    float or np.ndarray
        Energy shift of the Fermi level at the given theta_par value(s).
    """

    # Get EF_correction
    EF_fn = da.metadata.get_EF_correction()
    if isinstance(EF_fn, dict) and len(EF_fn) > 1:  # Theta-par-dep EF correction
        shift_str = ""
        for i in range(1, len(EF_fn)):
            shift_str += f"+ {str(EF_fn[f'c{i}'])} * theta_par ** {i}"
        if Ek is not None:
            shift_str += "+ Ek"
        return ne.evaluate(shift_str)
    else:
        if Ek is not None:
            return Ek
        else:
            return np.zeros_like(theta_par)


def _flatten_EF(da):
    """Removes curvature in the Fermi edge from a dispersion or other data.

    Parameters
    ----------
    da : xarray.DataArray
        Data to remove curvature from. Should have an EF_correction attribute set in the metadata.

    Returns
    -------
    xarray.DataArray
        Data with curvature removed.

    """
    # Get theta_par-dependent shift of EF
    EF_shift = _get_E_shift_at_theta_par(
        da,
        da.theta_par.data,
    )

    # Get new energy scale (ensure ordered low to high)
    EK_old = np.sort(da.eV.data)

    # Work out new energy scale
    dE = (EK_old[-1] - EK_old[0]) / (len(EK_old) - 1)
    EF_shift_min, EF_shift_max = np.min(EF_shift), np.max(EF_shift)
    if EF_shift_min < 0:
        EK_new = np.arange(EK_old[0], EK_old[-1] - EF_shift_min, dE)
    else:
        EK_new = np.arange(EK_old[0] - EF_shift_max, EK_old[-1], dE)

    # Calculate interpolation target
    Ek_values = EK_new[:, np.newaxis] + EF_shift[np.newaxis, :]
    theta_par_values = np.broadcast_to(da.theta_par.data[np.newaxis, :], Ek_values.shape)

    # Interpolate onto new energy scale
    interpolated_data = xr.apply_ufunc(
        _fast_bilinear_interpolate_rectilinear,
        Ek_values,
        theta_par_values,
        da.eV.data,
        da.theta_par.data,
        da.pint.dequantify(),
        input_core_dims=[
            ["eV", "theta_par"],
            ["eV", "theta_par"],
            ["eV"],
            ["theta_par"],
            ["eV", "theta_par"],
        ],
        output_core_dims=[["eV", "theta_par"]],
        exclude_dims={"eV"},
        vectorize=True,
        dask="parallelized",
        keep_attrs=True,
    )
    # Update co-ords for new EK values
    interpolated_data.coords.update({"eV": EK_new.flatten()})

    # Update metadata
    old_EF_correction_attrs = da.metadata.get_EF_correction()
    new_EF_correction_attrs = old_EF_correction_attrs["c0"]
    interpolated_data.metadata.calibration.set(
        "EF_correction", new_EF_correction_attrs, add_history=False
    )
    interpolated_data.history.add(
        f"Data interpolated to remove curavature of the Fermi edge from correction {old_EF_correction_attrs}. "
        f"New EF_correction set to {new_EF_correction_attrs}."
    )
    return interpolated_data.pint.quantify(eV=da.eV.units)


def _get_wf(da):
    """Gets the relevant work function to use for the data processing.

    Parameters
    ----------
    da : xarray.DataArray
        Data to extract the work function from.

    Returns
    -------
    wf : float or np.ndarray
        Work function to use for the data processing. Will be an array if an hv scan passed
    """

    # Get EF values, estimating if not already set
    Ek_at_EF = _get_EF_at_theta_par0(da)

    # Get photon energy
    if "hv" in da.dims:
        hv = da.hv.data
    else:
        hv = da.metadata.photon.hv
        if hv is not None:
            hv = hv.to("eV").magnitude
    if hv is None:
        hv = _add_estimated_hv(da)

    return hv - Ek_at_EF


def _get_BE_scale(da):
    """Gets the relevant binding energy scale from the kinetic energy scale, padding to account for any theta-par
    dependence of the Fermi level on the detector and any shift of EF within the detector for hv-dep data.

    Parameters
    ----------
    da : xarray.DataArray
        Data for extracting BE scale from

    Returns
    -------
    tuple
        (start, end, step) of the binding energy scale. Step is the same as the input data.
        Convention is used for negative values below Fermi level (i.e. :math:`\\omega=E-E_F` returned).
    """
    # Function implementation

    # Get current scales
    theta_par = da.theta_par.data
    # Work out theta_par-dep EF shift
    EF_shift = _get_E_shift_at_theta_par(da, theta_par)

    # Get current energy scale, taking a single hv if a hv scan
    if "hv" in da.dims:
        hv0 = da.hv.data[0]
        disp = da.disp_from_hv(hv0)
        EF_values = da.EF.sel(hv=hv0).data + EF_shift
    else:
        disp = da
        EF_values = _get_EF_at_theta_par0(da) + EF_shift
    eV = disp.eV.data
    eV_step = (eV[-1] - eV[0]) / (len(eV) - 1)

    # Work out new energy range, padding to avoid loosing real data at the edges due to theta_par dependence
    Emin, Emax = np.min(eV), np.max(eV)
    EF_min, EF_max = np.min(EF_values), np.max(EF_values)
    BE_min, BE_max = Emin - EF_max, Emax - EF_min

    # If hv scan, add extra padding to account for shift in EF on detector with hv
    if "hv" in da.dims:
        KE_shift_vs_hv = da.KE_delta.data - da.KE_delta.data[0]
        EF_shift_vs_hv = da.EF.data - da.EF.data[0]
        shift_on_detector = KE_shift_vs_hv - EF_shift_vs_hv
        if min(shift_on_detector) < 0:
            BE_max += np.abs(min(shift_on_detector))
        if max(shift_on_detector) > 0:
            BE_min -= max(shift_on_detector)

    return BE_min, BE_max + eV_step, eV_step


def _add_estimated_EF(da):
    """Adds estimated Fermi level to :class:`xarray.DataArray` attributes if existing EF correction is not present.

    Parameters
    ----------
    da : xarray.DataArray
        Data to add estimated Fermi level to.

    Returns
    -------
    None
        Adds the estimated Fermi level to the data attributes or coordinates as appropraite.

    """

    # Check no existing EF correction exists
    if da.metadata.calibration.EF_correction is not None:
        if "hv" in da.dims and "EF" not in da.coords:
            pass
        else:
            return

    # Estimate EF and add to data attributes
    estimated_EF = da.estimate_EF()
    if "hv" not in da.dims:
        da.metadata.set_EF_correction(estimated_EF)
        estimated_EF_str = f"{estimated_EF} eV"
    else:
        da.coords.update({"EF": ("hv", estimated_EF)})
        estimated_EF_str = (
            f"{estimated_EF[0]:.2f}-{estimated_EF[-1]:.2f} eV (hv dependent)"
        )

    # Update history and warning
    update_str = (
        f"EF_correction set from automatic estimation of Fermi level to: {estimated_EF_str}. "
        f"NB may not be accurate."
    )
    analysis_warning(update_str, "warning", "Analysis warning")
    da.history.add(update_str)


def _add_estimated_hv(da):
    """Adds estimated photon energy to :class:`xarray.DataArray` attributes if existing hv value is not set.

    Parameters
    ----------
    da : :class:`xarray.DataArray`
        Data to add estimated photon energy to.

    Returns
    -------
    None
        Adds the estimated photon energy to the data attributes.

    """

    # Check no existing hv value exists
    if "hv" in da.dims or (
        hasattr(hasattr(da.metadata, "photon"), "hv") and da.metadata.photon.hv
    ):
        return

    # Check if EF correction is present and estimate if not
    if da.metadata.calibration.EF_correction is None:
        _add_estimated_EF(da)
    EF = da.metadata.get_EF_correction()
    if isinstance(EF, dict):
        EF = EF["c0"]

    if EF > 16.5 and EF < 17:  # Likely to be He-I
        hv_est = 21.2182
        hv_est_str = str(hv_est) + " eV (He I)."
    elif EF > 1 and EF < 2:  # Likely to be 6.05 eV laser
        hv_est = 6.05
        hv_est_str = str(hv_est) + " eV (6 eV laser)."
    elif EF > 6 and EF < 7:  # Likely to be 11 eV laser
        hv_est = 10.897
        hv_est_str = str(hv_est) + " eV (11 eV laser)."
    else:  # Estimate from EF_est
        hv_est = EF + 4.4  # Taking 4.4 eV as a reasonable work function
        hv_est_str = str(hv_est) + " eV."

    da.metadata.photon.hv = hv_est

    # Update history and warning
    update_str = (
        f"hv set from automatic estimation of photon energy to: {hv_est_str}. "
        f"Check for accuracy."
    )
    analysis_warning(update_str, "warning", "Analysis warning")
    da.history.add(update_str)



import dask as da
import dill
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from scipy.ndimage import gaussian_filter1d
from scipy.signal import find_peaks
from tqdm.notebook import tqdm

from peaks.core.fitting.models import LinearDosFermiModel
from peaks.core.utils.misc import analysis_warning


def fit(
    data_array,
    model,
    params,
    independent_var=None,
    sequential=True,
    reverse_sequential_fit_order=False,
):
    """
    Fit an :class:`lmfit.Model` to an :class:`xarray.DataArray`, specifying the co-ordinate correspinding to
    the indepednent variable. Fit broadcasts along other dimensions. Returns the results of the fit as a
    :class:`xarray.DataArray` with uncertainties and the full :class:`lmfit.ModelResult`.

    Parameters:
    -----------
    data_array: xarray.DataArray
          The data to fit the model to.
    model: lmfit.Model
        The model to fit to the data.
    params: lmfit.Parameters
        Initial parameters for the model fitting.
    independent_var: str, optional
         The name of the dimension that represents the independent variable. If the data_array only has a single
         dimension, then this parameter can be omitted, but if the data_array has multiple dimensions, then it
         must be included.
    sequential: bool
        If True, and `data_array` is 2D, use the results of the previous fit to update the starting parameters
        for the next iteration. Assumes that the initial parameters are for the first slice
        (i.e. `data_array.isel(dim=0)` slice) along the non-independent dimension.
        Defaults to True.
        If False, fit the model to the entire data array at once.
    reverse_sequential_fit_order: bool
        Use to reverse the order of a sequential fit along the non-independent dimension. Defaults to False.

    Returns:
    -------
    xarray.DataSet
        A DataSet containing the best-fit parameters, their uncertainties, and the :class:`lmfit.ModelResult` object.
    """

    def fit_func(y, x, model, initial_params):
        result = model.fit(y, params=initial_params, x=x)
        best_values = np.array([result.params[param].value for param in result.params])
        uncertainties = np.array(
            [
                (
                    result.params[param].stderr
                    if result.params[param].stderr is not None
                    else np.nan
                )
                for param in result.params
            ]
        )
        return np.concatenate([best_values, uncertainties, [result]])

    # Dequantify the data array
    data_array = data_array.pint.dequantify()

    # Check independent_var is a valid dimension if supplied and set if None
    if independent_var is not None:
        if independent_var not in data_array.dims:
            raise ValueError(
                f"Independent variable {independent_var} is not a valid dimension of the data array."
                f"Expected one of {data_array.dims}"
            )
    else:
        if data_array.ndim != 1:
            raise ValueError(
                "Data array has multiple dimensions. Please specify the independent variable using the "
                "`independent_var` argument."
            )
        else:
            independent_var = data_array.dims[0]

    if data_array.ndim > 2 and sequential:
        sequential = False
        analysis_warning(
            "Sequential fitting only supported for 2D data. Defaulting to non-sequential.",
            title="Analysis info",
            warn_type="info",
        )
    # Sequential fitting, updating params each iteration
    if sequential and len(data_array.dims) == 2:
        if isinstance(data_array.data, da.array.core.Array):
            raise ValueError(
                "Dask arrays are not supported for sequential fitting. Either set `sequential=False` or load your "
                "data into memory first with ``.compute()``"
            )

        non_indep_dim = list(set(data_array.dims) - set([independent_var]))[0]
        non_indep_dim_len = data_array.sizes[non_indep_dim]

        if reverse_sequential_fit_order:
            # Reverse the order of the non-independent dimension
            data_array = data_array.isel({non_indep_dim: slice(None, None, -1)})

        fit_results = []
        # Iterate through all slices along the non-independent dimension
        for i in tqdm(range(non_indep_dim_len), desc="Fitting"):
            data_array_slice = data_array.isel({non_indep_dim: i})
            results_subset = xr.apply_ufunc(
                fit_func,
                data_array_slice,
                data_array.coords[independent_var],
                kwargs={"model": model, "initial_params": params},
                input_core_dims=[[independent_var], [independent_var]],
                output_core_dims=[["fit_params"]],
                vectorize=True,
                output_dtypes=[object],
                keep_attrs=False,
            )
            fit_results.append(results_subset)

            # Update the initial parameters for the next iteration
            params = results_subset.isel(fit_params=-1).item().params
        # Concatenate the results along the non-independent dimension into a single DataArray
        results = xr.concat(fit_results, dim=non_indep_dim)
    else:
        # Apply the fitting function across all dimensions except the independent variable
        results = xr.apply_ufunc(
            fit_func,
            data_array,
            data_array.coords[independent_var],
            kwargs={"model": model, "initial_params": params},
            input_core_dims=[[independent_var], [independent_var]],
            output_core_dims=[["fit_params"]],
            vectorize=True,
            dask="parallelized",
            output_dtypes=[object],
            dask_gufunc_kwargs={
                "output_sizes": {"fit_params": len(params) * 2 + 1},
                "allow_rechunk": True,
            },
            keep_attrs=False,
        )

    # Create parameter names, adding "_stderr" for uncertainties, and "model_result" for the serialized data
    param_names = list(params.keys())
    all_param_names = (
        param_names + [f"{name}_stderr" for name in param_names] + ["fit_model"]
    )

    # Add the parameter names to the dataarray
    results = results.assign_coords({"fit_params": ("fit_params", all_param_names)})

    # Parse these as a dataset
    results_ds = xr.Dataset()

    # Iterate over all_param_names and assign each corresponding value from results to the new dataset,
    # ensuring dtype of float if not the fit model
    for i, param_name in enumerate(all_param_names):
        data_var = results.isel(fit_params=i).drop_vars("fit_params")
        if param_name != "fit_model":
            data_var = data_var.astype(np.float64)
        results_ds[param_name] = data_var

    results_ds.attrs["independent_var"] = independent_var
    return results_ds


def fit_gold(data, EF_correction_type="poly4", **kwargs):
    """
    Helper function for fitting a gold reference scan to a standard LinearDosFermiModel with parameters:
    - Fermi level (EF)
    - Temperature (T)
    - DOS slope (dos_slope)
    - DOS intercept (dos_intercept)
    - Background slope (bg_slope)
    - Background intercept (bg_intercept)
    - Gaussian convolution (sigma_conv)

    Parameters:
    -----------
    data_array : xarray.DataArray
        The gold reference data to fit.

    EF_correction_type : str, optional
        The type of Fermi level correction to determine for 2D data. Options are:
        - 'poly4' (default): Fit a 4th order polynomial to the extracted Fermi level from all slices along the
        non-independent dimension.
        - 'poly3': Fit a 3rd order polynomial to the extracted Fermi level values.
        - 'quadratic': Fit a quadratic to the extracted Fermi level values.
        - 'linear': Fit a linear function to the extracted Fermi level values.
        - 'average': Average the extracted Fermi level values from all slices

    **kwargs : optional
        Additional keyword arguments to initialise paramaeter values

    Returns:
    -------
    xarray.DataSet
        A DataSet containing the best-fit parameters, their uncertainties, and the :class:`lmfit.ModelResult` object.

    Examples:
    --------
    Example usage is as follows::

        import peaks as pks

        # Load the gold reference data
        gold_data = pks.load(gold_scan)

        # Fit the gold reference data, initialising the background slope to 0
        gold_fit = pks.fit_gold(gold_data, bg_slope=0)

    """
    data = data.pint.dequantify()
    if data.ndim > 2:
        raise ValueError("Expected 1D or 2D DataArray to be supplied.")
    if "eV" not in data.dims:
        raise ValueError("Supplied data must include a dimension of 'eV'")

    gold_model = LinearDosFermiModel()
    if data.ndim == 1:
        params = gold_model.guess(data, **kwargs)
        fit_result = data.fit(gold_model, params)
        fit_result.plot_fit(show_components=False)
        results_text = (
            f"Fermi energy: {str(np.round(fit_result['EF'].data, 3))} ± "
            f"{str(np.round(fit_result['EF_stderr'].data, 3))} eV,  "
            f"Resolution: {str(np.round(fit_result['sigma_conv'].data, 3))} ± "
            f"{str(np.round(fit_result['sigma_conv_stderr'].data, 3))} eV"
        )
        fit_result.attrs["EF_correction"] = float(fit_result["EF"])

    elif data.ndim == 2:
        other_dim = list(set(data.dims) - set(["eV"]))[0]
        first_slice = data.isel({other_dim: 0})
        params = gold_model.guess(first_slice, **kwargs)
        fit_result = data.fit(gold_model, params, independent_var="eV")

        # Fit the Fermi level correction
        if EF_correction_type == "poly4":
            _order = 4
        elif EF_correction_type == "poly3":
            _order = 3
        elif EF_correction_type == "quadratic":
            _order = 2
        elif EF_correction_type == "linear":
            _order = 1
        elif EF_correction_type == "average":
            _order = 0
        EF_correction_fit = fit_result["EF"].quick_fit.poly(_order)

        if EF_correction_type == "average":
            EF_correction = float(EF_correction_fit["c0"])
            results_text = (
                f"Average Fermi level correction: {np.round(EF_correction, 3)}"
            )
        else:
            EF_correction = {
                f"c{i}": float(EF_correction_fit[f"c{i}"]) for i in range(_order + 1)
            }
            results_text = (
                f"Polynomial Fermi level correction (order {_order}): "
                + ", ".join(
                    [f"{key}: {np.round(val, 6)}" for key, val in EF_correction.items()]
                )
            )

        fit_result.attrs["EF_correction"] = EF_correction
        res_av = np.round(fit_result["sigma_conv"].mean().data, 3)
        results_text += f";   Average resolution: {res_av} eV"

        # Make some plots
        fig = plt.figure(figsize=(8, 5))
        gs = fig.add_gridspec(3, 2, width_ratios=[1, 1], height_ratios=[1, 2, 2])
        ax1 = fig.add_subplot(gs[:, 0])
        data.plot(ax=ax1, add_colorbar=False)
        fit_result["EF"].plot(ax=ax1)
        ax1.set_ylabel("eV")

        ax2 = fig.add_subplot(gs[0, 1])
        EF_correction_fit.fit_model.data[()].plot_residuals(ax=ax2)
        ax2.xaxis.set_visible(False)
        ax3 = fig.add_subplot(gs[1, 1], sharex=ax2)
        EF_correction_fit.fit_model.data[()].plot_fit(ax=ax3)
        ax3.xaxis.set_visible(False)
        ax3.set_title("")
        ax3.set_ylabel("Fermi level (eV)")
        ax4 = fig.add_subplot(gs[2, 1], sharex=ax2)
        fit_result["sigma_conv"].plot(ax=ax4)
        ax4.set_title("Resolution")
        ax4.axhline(res_av, color="r", ls="--")

        plt.tight_layout()
        plt.show()

    analysis_warning(results_text, "success", "Au fitting results")
    return fit_result


def _estimate_EF(y, x):
    """
    Estimate the Fermi level from the first negative peak in the first derivative of the data.

    Parameters:
    -----------
    y : numpy.ndarray
        The data array to estimate the Fermi level from.
    x : numpy.ndarray
        The corresponding x-axis values.

    Returns:
    --------
    float
        The estimated Fermi level.
    """

    def noise_level(y):
        # Calculate the average of nearest neighbour differences to evaluate the noise level of a signal
        NNdist = [abs(y[i] - y[i - 1]) for i in range(1, len(y))]
        noise = np.average(NNdist)
        return noise

    # Smooth the data and compute its derivative
    sigma_px = 0.01 / (abs(x[-1] - x[0]) / len(x))
    deriv = np.gradient(gaussian_filter1d(y, sigma=sigma_px), x)
    y_filtered = -gaussian_filter1d(deriv, 2)
    noise = noise_level(y_filtered)

    # Find all the peaks with prominence >= 2.5 * noise level and with width at least 3 points
    peaks_index, _ = find_peaks(y_filtered, prominence=noise * 2.5, width=3)
    EF = np.round(x[peaks_index].max(), 3)  # Estimated EF
    return EF


def estimate_EF(da):
    """Make an approximate guess for the Fermi level from the corresponding peak in the derivative of the data
    :::{warning}
    This is only a very approximate method for use in making estiamtes to feed into fit functions and GUIs etc.
    It should not be used as a true determination of the Fermi level.
    :::

    Parameters:
    -----------
    data: xarray.DataArray
        The data to estimate the Fermi level from.

    Returns:
    --------
    float
        The estimated Fermi level.
    """

    if "eV" not in da.dims:
        raise ValueError("Data must have an 'eV' dimension to estimate the Fermi level.")

    # Check for an hv scan
    if "hv" in da.dims and "kinetic" in da.metadata.analyser.scan.eV_type.lower():
        # Iterate through the photon energies and estimate at each
        EF_values = []
        for hv in da.hv.data:
            data_hv = da.disp_from_hv(hv=hv)
            EF = _estimate_EF(data_hv.DOS().fillna(0).data, data_hv.eV.data)
            EF_values.append(EF)
        EF_data = xr.DataArray(EF_values, dims=["hv"], coords={"hv": da.hv.data})
        # Fit the result to a 2nd order polynomial
        fit_order = 3
        fit_result = EF_data.quick_fit.poly(fit_order)
        fit_model = fit_result.fit_model.data[()]
        params = {
            f"c{i}": f"{fit_model.params[f'c{i}'].value:.5f}"
            for i in range(fit_order + 1)
        }
        analysis_warning(
            f"Estimated Fermi level correction across the photon energy range via an order {fit_order} polynomial fit as {params}",
            "info",
            "Analysis info",
        )
        fit_result.plot_fit(ylabel="$E_\\mathrm{F}$ (eV)", xlabel="$h\\nu$ (eV)")
        # Back-calculate the Fermi level from the fit at each photon energy
        EF_values_out = fit_model.eval(x=EF_data.hv.data)
        return EF_values_out
    else:
        try:
            return _estimate_EF(da.DOS().fillna(0).pint.dequantify().data, da.eV.data)
        except Exception:
            return None


def save_fit(fit_result, filename):
    """
    Save the results of a fit.

    Parameters:
    -----------
    fit_result : xarray.DataSet
        The results of the fit to save.
    filename : str
        The name of the file to save the fit results to.

    Returns:
    --------
    None
    """

    result_as_dict = fit_result.to_dict()
    with open(filename, "wb") as f:
        dill.dump(result_as_dict, f, protocol=-1)


def load_fit(filename):
    """
    Load the results of a fit from a netCDF file.

    Parameters:
    -----------
    filename : str
        The name of the file to load the fit results from.

    Returns:
    --------
    xarray.DataSet
        The results of the fit.
    """

    with open(filename, "rb") as f:
        result_as_dict = dill.load(f)
    return xr.Dataset.from_dict(result_as_dict)
